#coding:utf-8
import os
import shutil
import tarfile
import zipfile

from six.moves import urllib


class Folder(object):
    def __init__(self, from_path):
        self.from_path = from_path

    def extractall(self, path):
        os.rmdir(path)
        shutil.copytree(self.from_path, path)

    def close(self):
        pass


def open_zipfile(archive):
    return zipfile.ZipFile(archive)


def open_tarfile(archive):
    return tarfile.open(name=archive, mode='r')


SUPPORTED_ARCHIVES = [
    (open_tarfile, tarfile.is_tarfile),
    (open_zipfile, zipfile.is_zipfile),
    (Folder, os.path.isdir),
]


def bootstrap_from_archive(path, skeleton):
    if os.path.isfile(skeleton) or os.path.isdir(skeleton):
        # Is is a local file?
        skeleton_file = skeleton
    else:
        # Assume it's an URL
        skeleton_file, headers = urllib.request.urlretrieve(skeleton)

    for opener, test in SUPPORTED_ARCHIVES:
        try:
            if test(skeleton_file):
                archive = opener(skeleton_file)
                break
        except IOError:
            pass
    else:
        raise Exception("Unsupported skeleton file type. Only .tar and .zip are supported at this time.")

    os.mkdir(path)
    archive.extractall(path=path)
    archive.close()



import os
import os.path
import shutil
import tempfile


has_symlink = False


compat_test_dir = tempfile.mkdtemp()

# Check for symlink support (available and usable)

src = os.path.join(compat_test_dir, "src")
dst = os.path.join(compat_test_dir, "dst")

with open(src, "w"):
    pass

try:
    os.symlink(src, dst)
except (AttributeError, OSError):
    # AttributeError if symlink is not available (Python <= 3.2 on Windows)
    # OSError if we don't have the symlink privilege (on Windows)
    pass  # Leave has_symlink false
else:
    has_symlink = True


shutil.rmtree(compat_test_dir)



#coding:utf-8
import os
import logging

from cactus.utils.url import path_to_url


logger = logging.getLogger(__name__)


class PageContextCompatibilityPlugin(object):
    """
    This plugin ensures that the page context stays backwards compatible, but adds
    deprecation warnings.
    """
    def preBuildPage(self, page, context, data):
        prefix = os.path.relpath(".", os.path.dirname(page.build_path))

        def static_url():
            logger.warn("%s:", page.path)
            logger.warn("{{ STATIC_URL }} is deprecated, use {% static '/static/path/to/file' %} instead.")
            return path_to_url(os.path.join(prefix, 'static'))

        def root_url():
            logger.warn("%s:", page.path)
            logger.warn("{{ ROOT_URL }} is deprecated, use {% url '/page.html' %} instead.")
            return path_to_url(prefix)

        def page_url():
            logger.warn("%s:", page.path)
            logger.warn("{{ PAGE_URL }} is deprecated, use {% current_page %} instead")
            return page.final_url[1:]  # We don't want the leading slash (backwards compatibility)

        context.update({
            "STATIC_URL": static_url,
            "ROOT_URL": root_url,
            "PAGE_URL": page_url,
        })

        return context, data



#coding:utf-8
import os
import time
import threading
import logging

import six

from cactus.utils.filesystem import fileList
from cactus.utils.network import retry

class PollingListener(object):
    def __init__(self, path, f, delay = .5, ignore = None):
        self.path = path
        self.f = f
        self.delay = delay
        self.ignore = ignore
        self._pause = False
        self._checksums = {}

    def checksums(self):
        checksumMap = {}

        for f in fileList(self.path):

            if f.startswith('.'):
                continue

            if self.ignore and self.ignore(f):
                continue

            try:
                checksumMap[f] = int(os.stat(f).st_mtime)
            except OSError:
                continue

        return checksumMap

    def run(self):
        # self._loop()
        t = threading.Thread(target=self._loop)
        t.daemon = True
        t.start()

    def pause(self):
        self._pause = True

    def resume(self):
        self._checksums = self.checksums()
        self._pause = False

    def _loop(self):
        self._checksums = self.checksums()

        while True:
            self._run()

    @retry((Exception,), tries = 5, delay = 0.5)
    def _run(self):
        if not self._pause:
            oldChecksums = self._checksums
            newChecksums = self.checksums()

            result = {
                'added': [],
                'deleted': [],
                'changed': [],
            }

            for k, v in six.iteritems(oldChecksums):
                if k not in newChecksums:
                    result['deleted'].append(k)
                elif v != newChecksums[k]:
                    result['changed'].append(k)

            for k, v in six.iteritems(newChecksums):
                if k not in oldChecksums:
                    result['added'].append(k)

            result['any'] = result['added'] + result['deleted'] + result['changed']

            if result['any']:
                self._checksums = newChecksums
                self.f(result)

        time.sleep(self.delay)



#coding:utf-8
from cactus.tests.deployment import DummyUI, DummySite, DummyDeploymentEngine, BaseDeploymentTestCase


class BucketCreateTestCase(BaseDeploymentTestCase):
    def setUp(self):
        super(BucketCreateTestCase, self).setUp()
        self.ui = DummyUI()
        self.site = DummySite(self.test_dir, self.ui)
        self.engine = DummyDeploymentEngine(self.site)
        self.engine.configure()

    def test_bucket_attrs(self):
        """
        Test that the bucket name is provided
        """
        self.assertEqual("test-bucket", self.engine.bucket_name)
        self.assertEqual("test-bucket-obj", self.engine.bucket)

    def test_config_saved(self):
        """
        Test that the configuration is saved
        """
        self.assertEqual("test-bucket", self.site.config.get("test-conf-entry"))
        self.assertEqual("http://test-bucket.com", self.site.config.get("test-conf-entry-website"))

    def test_credentials_saved(self):
        """
        Test that the credentials are saved
        """
        self.assertTrue(self.engine.credentials_manager.saved)



#coding:utf-8
import os
import getpass
import keyring

is_desktop_app = os.environ.get("DESKTOPAPP", None) not in ["", None]

def get_password(service, account):

    # Because we cannot use keychain from a sandboxed app environment we check if the password
    # was passed by the app into the env.
    if is_desktop_app:
        return os.environ.get("SECRET_KEY", None)

    return keyring.get_password(service, account)

def set_password(service, account, password):

    if is_desktop_app:
        return

    keyring.set_password(service, account, password)


class BaseKeyringCredentialsManager(object):
    _username_config_entry = "username"
    _keyring_service = "cactus"

    _username_display_name = "Username"
    _password_display_name = "Password"

    def __init__(self, engine):
        self.engine = engine  #TODO: Don't we want only UI and config?
        self.username = None
        self.password = None

    def get_credentials(self):
        self.username = self.engine.site.config.get(self._username_config_entry)
        if self.username is None:
            self.username = self.engine.site.ui.prompt("Enter your {0}".format(self._username_display_name))

        self.password = get_password(self._keyring_service, self.username)

        if self.password is None:
            self.password = self.engine.site.ui.prompt("Enter your {0} (will not be echoed)".format(self._password_display_name),
                prompt_fn=getpass.getpass)

        return self.username, self.password

    def save_credentials(self):
        assert self.username is not None, "You did not set {0}".format(self._username_display_name)
        assert self.password is not None, "You did not set {0}".format(self._password_display_name)

        self.engine.site.config.set(self._username_config_entry, self.username)
        self.engine.site.config.write()

        set_password(self._keyring_service, self.username, self.password)



#!/usr/bin/env python
# encoding: utf-8
import os
import sys
import time
import argparse
import socket

from six.moves import input
import colorama

# No cactus imports here! There's no logging in place (or anything really).
# It's best to wait until we've ran setup_logging before running imports (some import soft errors log).


class CactusCli(object):
    """
    We wrap all imports into this object to control their order
    """
    def __init__(self):
        self.Site = None
        self.bootstrap = None

    def do_imports(self):
        from cactus.site import Site
        from cactus.bootstrap import bootstrap
        self.Site = Site
        self.bootstrap = bootstrap

    def create(self, path, skeleton=None):
        """
        Creates a new project at the given path.

        :param path: The path where the new project should be created
        :param skeleton: An (optional) skeleton to use to create the project.
                         This could be a zip, tar archive (file path or URL)
        """
        if os.path.exists(path):
            new_path = '%s.%s.moved' % (path, int(time.time()))
            if input('Path %s exists, move aside to %s? (y/n) ' % (path, new_path)) == 'y':
                os.rename(path, new_path)
            else:
                sys.exit()

        self.bootstrap(path, skeleton)

    def build(self, path, config):
        """Build a cactus project"""
        site = self.Site(path, config, verb=self.Site.VERB_BUILD)
        site.build()

    def deploy(self, path, config):
        """Upload the project to S3"""
        site = self.Site(path, config, verb=self.Site.VERB_DEPLOY)
        site.upload()

    def make_messages(self, path, config):
        """ Create the list of translation files for the site """
        site = self.Site(path, config)
        site.make_messages()

    def serve(self, path, config, port, browser):
        """Serve the project and watch changes"""
        site = self.Site(path, config, verb=self.Site.VERB_SERVE)
        site.serve(port=port, browser=browser)

    def domain_setup(self, path, config):
        site = self.Site(path, config)
        site.domain_setup()

    def domain_list(self, path, config):
        site = self.Site(path, config)
        site.domain_list()


def parse_arguments(cli, args):
    """Parse command line arguments"""

    parser = argparse.ArgumentParser(description="Build and deploy static websites using Django templates.")

    subparsers = parser.add_subparsers(title='subcommands', description='Valid subcommands',
                                       help='Select a command to run.', dest='command')
    subparsers.required = True

    parser_create = subparsers.add_parser('create', help='Create a new project')
    parser_create.add_argument('path', help='The path where the new project should be created')
    parser_create.add_argument('-s', '--skeleton', help='An archive to use as skeleton to create the new project')
    parser_create.set_defaults(target=cli.create)

    parser_build = subparsers.add_parser('build', help='Build the current project.')
    parser_build.set_defaults(target=cli.build)

    parser_deploy = subparsers.add_parser('deploy', help='Deploy the current project to S3.')
    parser_deploy.set_defaults(target=cli.deploy)

    parser_serve = subparsers.add_parser('serve', help='Serve the current project.')
    parser_serve.set_defaults(target=cli.serve)
    parser_serve.add_argument('-p', '--port', default=8000, type=int, help='The port on which to serve the site.')
    parser_serve.add_argument('-b', '--browser', action='store_true', help='Whether to open a browser for the site.')

    parser_make_messages = subparsers.add_parser('messages:make', help='Create translation files for the current project')
    parser_make_messages.set_defaults(target=cli.make_messages)

    parser_domain_setup = subparsers.add_parser('domain:setup', help='Setup records for a domain with route 53')
    parser_domain_setup.set_defaults(target=cli.domain_setup)

    parser_domain_list = subparsers.add_parser('domain:list', help='Setup records for a domain with route 53')
    parser_domain_list.set_defaults(target=cli.domain_list)


    config_parsers = [parser_build, parser_deploy, parser_serve, parser_make_messages, parser_domain_setup, parser_domain_list]
    all_parsers = config_parsers + [parser_create]

    for subparser in config_parsers:
        subparser.add_argument('-c', '--config', action="append",
                               help='Add a config file you want to use')
        subparser.add_argument('-d', '--path', default=os.getcwd(),
                               help='The path to the Cactus project')

    for subparser in all_parsers:
        verbosity_group = subparser.add_mutually_exclusive_group()
        verbosity_group.add_argument('-v', '--verbose', action='store_true', help='Be more verbose')
        verbosity_group.add_argument('-q', '--quiet', action='store_true', help='Be quieter')

    ns = parser.parse_args(args)

    # Small hack to provide a default value while not replacing what's
    # given by the user, if there is
    if hasattr(ns, 'config') and ns.config is None:  # We don't need config for create
        ns.config = [os.path.join(ns.path, 'config.json')]

    return ns

def main(args):
    cli = CactusCli()

    # CLI parsing
    ns = parse_arguments(cli, args)

    # Colors!
    colorama.init()

    # Logging
    from cactus.logger import setup_logging
    setup_logging(ns.verbose, ns.quiet)

    # Network setup that we should presumably remove. Leaving it there for now:
    # it's a better place than cactus/__init__.py
    socket.setdefaulttimeout(5)

    # Import Cactus packages and run required command.
    cli.do_imports()

    kwargs = dict((k, v) for k, v in vars(ns).items() if k not in ['command', 'target', 'verbose', 'quiet'])
    ns.target(**kwargs)


def cli_entrypoint():
    main(sys.argv[1:])


if __name__ == "__main__":
    cli_entrypoint()



"""
Modify `config.json` to set a custom blog path, default author name, or date pattern used to parse metadata. The defaults are:
"blog": {
    "path": "blog",
    "author": "Unknown",
    "date-format": "%d-%m-%Y"
}
"""

import os
import datetime
import logging

ORDER = 999
POSTS = []

from django.template import Context
from django.template.loader import get_template
from django.template.loader_tags import BlockNode, ExtendsNode

def getNode(template, context=Context(), name='subject'):
    """
    Get django block contents from a template.
    http://stackoverflow.com/questions/2687173/
    django-how-can-i-get-a-block-from-a-template
    """
    for node in template:
        if isinstance(node, BlockNode) and node.name == name:
            return node.render(context)
        elif isinstance(node, ExtendsNode):
            return getNode(node.nodelist, context, name)
    raise Exception("Node '%s' could not be found in template." % name)


def preBuild(site):

    global POSTS
    siteContext = site.context()
    
    blog = site.config.get('blog', {})
    blogPath = os.path.join(blog.get('path', 'blog'), '')
    dateFormat = blog.get('date-format', '%d-%m-%Y')
    defaultAuthor = blog.get('author', 'Unknown')

    # Build all the posts
    for page in site.pages():
        if page.path.startswith(blogPath):

            # Skip non html posts for obious reasons
            if not page.path.endswith('.html'):
                continue

            # Find a specific defined variable in the page context,
            # and throw a warning if we're missing it.
            def find(name):
                c = page.context()
                if not name in c:
                    logging.info("Missing info '%s' for post %s" % (name, page.path))
                    return ''
                return c.get(name, '')

            # Build a context for each post
            context = {'__CACTUS_CURRENT_PAGE__': page,}
            context.update(siteContext)

            postContext = {}
            postContext['title'] = find('title')
            postContext['author'] = find('author') or defaultAuthor
            postContext['date'] = find('date')
            postContext['path'] = page.final_url
            postContext['body'] = getNode(get_template(page.path), context=Context(context), name="body")

            # Parse the date into a date object
            try:
                postContext['date'] = datetime.datetime.strptime(postContext['date'], dateFormat)
            except Exception as e:
                logging.warning("Date format not correct for page %s, should be %s\n%s" % (page.path, dateFormat, e))
                continue

            POSTS.append(postContext)

    # Sort the posts by date
    POSTS = sorted(POSTS, key=lambda x: x['date'])
    POSTS.reverse()

    indexes = xrange(0, len(POSTS))

    for i in indexes:
        if i+1 in indexes: POSTS[i]['prevPost'] = POSTS[i+1]
        if i-1 in indexes: POSTS[i]['nextPost'] = POSTS[i-1]


def preBuildPage(site, page, context, data):
    """
    Add the list of posts to every page context so we can
    access them from wherever on the site.
    """
    context['posts'] = POSTS

    for post in POSTS:
        if post['path'] == page.final_url:
            context.update(post)

    return context, data



import os
import shutil
import tempfile
from contextlib import contextmanager

from cactus.utils.helpers import map_apply


def mkdtemp():
    return tempfile.mkdtemp(dir=os.environ.get("TEMPDIR"))

def fileList(paths, relative=False, folders=False):
    """
    Generate a recursive list of files from a given path.
    """

    if not isinstance(paths, list):
        paths = [paths]

    files = []

    def append(directory, name):
        if not name.startswith('.'):
            path = os.path.join(directory, name)
            files.append(path)

    for path in paths:
        for directory, dirnames, filenames in os.walk(path, followlinks=True):
            if folders:
                for dirname in dirnames:
                    append(directory, dirname)
            for filename in filenames:
                append(directory, filename)
        if relative:
            files = map_apply(lambda x: x[len(path) + 1:], files)

    return files


@contextmanager
def alt_file(current_file):
    """
    Create an alternate file next to an existing file.
    """
    _alt_file = current_file + '-alt'
    yield _alt_file
    try:
        shutil.move(_alt_file, current_file)
    except IOError:
        # We didn't use an alt file.
        pass


@contextmanager
def chdir(new_dir):
    """
    Chdir to another directory for an operation
    """
    current_dir = os.getcwd()
    os.chdir(new_dir)
    yield
    os.chdir(current_dir)



#coding:utf-8
from __future__ import print_function
from six.moves import urllib
from six.moves import input


class InvalidInput(Exception):
    """
    Raised when an input is invalid.
    You should use this in your `coerce_fn` passed to `prompt`, to indicate that the
    input you received was no valid.
    """
    def __init__(self, reason=""):
        self.reason = reason


def prompt(q, coerce_fn=None, error_msg="Invalid input, please try again", prompt_fn=input):
    """
    :param q: The prompt to display to the user
    :param coerce_fn: A function to coerce, and validate, the user input.
        You may raise InvalidInput in it, to indicate that the input was not valid.
    :param error_msg: An error message to display if the input is incorrect
    :return: The user's input
    :rtype: str
    """
    if coerce_fn is None:
        coerce_fn = lambda x:x

    while 1:
        r = prompt_fn(q + " > ")
        try:
            return coerce_fn(r)
        except InvalidInput as e:
            print(e.reason or error_msg)


_yes_no_mapping = {"y":True, "n":False}
def _yes_no_coerce_fn(r):
    """
    :rtype: bool
    """
    try:
        return _yes_no_mapping[r.lower()]
    except KeyError:
        raise InvalidInput("Please enter `y` or `n`")

def prompt_yes_no(q):
    """
    :param q: The prompt to display to the user
    :return: True of False
    :rtype : bool
    """
    return prompt(q + " [y/n]", _yes_no_coerce_fn)


def _normalized_coerce_fn(r):
    """
    :rtype: str
    """
    return r.lower().strip()

def prompt_normalized(q):
    """
    :param q: The prompt to display to the user
    :returns: The user's normalized input
    :rtype: str
    """
    return prompt(q, _normalized_coerce_fn)


def _url_coerce_fn(r):
    """
    :rtype: str
    """
    p = urllib.parse.urlparse(r)
    if not p.scheme:
        raise InvalidInput("Specify an URL scheme (e.g. http://)")
    if not p.netloc:
        raise InvalidInput("Specify a domain (e.g. example.com)")
    if p.path and p.path != "/":
        raise InvalidInput("Do not specify a path")
    if p.params or p.query or p.fragment:
        raise InvalidInput("Do not leave trailing elements")

    if not p.path:
        r += "/"  #TODO: Fixme once the sitemap is fixed!
    r = r.lower()

    return r


def prompt_url(q):
    """
    :param q: The prompt to display to the user
    :return: The user's normalized input. We ensure there is an URL scheme, a domain, a "/" path,
        and no trailing elements.
    :rtype: str
    """
    return prompt(q, _url_coerce_fn)



#coding:utf-8
from cactus.tests import SiteTestCase


class TestCompatibility(SiteTestCase):
    def _paths_key_exists(self, obj, old_key):
        try:
            obj.paths[old_key]
        except KeyError:
            self.fail('Old key does not exist anymore: {0}'.format(old_key))

    def test_compatibility(self):
        """
        Test that we can access the path elements the "old" way
        Just try a few of them, we're not locking this down - just testing
        """
        for old_key, new_field in (
            ('build', 'build_path'),
            ('pages', 'page_path'),
        ):
            self.assertEqual(self.site.paths[old_key], getattr(self.site, new_field))

    def test_site_paths_keys_exist(self):
        for old_key in ('build', 'pages', 'plugins', 'templates', 'static', 'script'):
            self._paths_key_exists(self.site, old_key)

    def test_page_paths_keys_exist(self):
        page = self.site.pages()[0]

        for old_key in ('full', 'full-build'):
            self._paths_key_exists(page, old_key)

    def test_page_path_attr(self):
        page = self.site.pages()[0]
        self.assertEqual(page.source_path, page.path)

    def test_page_paths_keys_exist_in_static(self):
        static = self.site.static()[0]

        for old_key in ('full', 'full-build'):
            self._paths_key_exists(static, old_key)



class ExternalManager(object):
    """
    Manager the active externals
    """
    def __init__(self, site, processors=None, optimizers=None):
        self.site = site
        self.processors = processors if processors is not None else []
        self.optimizers = optimizers if optimizers is not None else []

    def _register(self, external, externals):
        externals.insert(0, external)

    def _deregister(self, external, externals):
        externals.remove(external)

    def clear(self):
        """
        Clear this manager
        """
        self.processors = []
        self.optimizers = []

    def register_processor(self, processor):
        """
        Add a new processor to the list of processors
        This processor will be added with maximum priority
        """
        self._register(processor, self.processors)

    def deregister_processor(self, processor):
        """
        Remove an existing processor from the list
        Will raise a ValueError if the processor is not present
        """
        self._deregister(processor, self.processors)

    def register_optimizer(self, optimizer):
        """
        Add a new optimizer to the list of optimizer
        This optimizer will be added with maximum priority
        """
        self._register(optimizer, self.optimizers)

    def deregister_optimizer(self, processor):
        """
        Remove an existing optimizer from the list
        Will raise a ValueError if the optimizer is not present
        """
        self._deregister(processor, self.optimizers)



# #coding:utf-8
# import sys
# import os
# import subprocess
# import time
#
# import requests
# from requests.adapters import HTTPAdapter
# from requests.packages.urllib3 import Retry
#
# from cactus.tests import BaseTestCase
# from cactus.utils.filesystem import fileList
#
#
# # Python < 3.3 compatibility
# if not hasattr(subprocess, "DEVNULL"):
#     subprocess.DEVNULL = open(os.devnull, 'w')
#
#
# class CliTestCase(BaseTestCase):
#     def find_cactus(self):
#         """
#         Locate a Cactus executable and ensure that it'll run using the right interpreter.
#         This is meant to work well in a Tox environment. That's how we run our tests, so that's all that really
#         matters here.
#         """
#         bin_path = os.path.abspath(os.path.dirname(sys.executable))
#
#         try:
#             path = os.path.join(bin_path, "cactus")
#             with open(path) as f:
#                 self.assertEqual("#!{0}".format(sys.executable), f.readline().strip())
#         except IOError:
#             pass
#         else:
#             return path
#
#         try:
#             path = os.path.join(bin_path, "cactus.exe")
#             with open(path):
#                 pass
#         except IOError:
#             pass
#         else:
#             return path
#
#         self.fail("Unable to find Cactus")
#
#     def run_cli(self, args, stdin="", cwd=None):
#         real_args = [self.find_cactus()]
#         real_args.extend(args)
#
#         kwargs = {
#             "args": real_args,
#             "stdin": subprocess.PIPE,
#             "stdout": subprocess.PIPE,
#             "stderr": subprocess.PIPE,
#         }
#
#         if cwd is not None:
#             kwargs["cwd"] = cwd
#
#         p = subprocess.Popen(**kwargs)
#         out, err = p.communicate(stdin.encode("utf-8"))
#         return p.returncode, out.decode("utf-8"), err.decode("utf-8")
#
#     def test_create_and_build(self):
#         self.assertEqual(0, len(os.listdir(self.test_dir)))
#
#         # Test regular create
#         ret, out, err = self.run_cli(["create", self.path, "--skeleton", os.path.join("cactus", "tests", "data", "skeleton")])
#         self.assertEqual(0, ret)
#
#         self.assertEqual(
#             sorted(fileList(self.path, relative=True)),
#             sorted(fileList("cactus/tests/data/skeleton", relative=True)),
#         )
#
#         # Test that we prompt to move stuff out of the way
#         _, _, _ = self.run_cli(["create", "-v", self.path], "n\n")
#         self.assertEqual(1, len(os.listdir(self.test_dir)))
#
#         _, _, _ = self.run_cli(["create", "-q", self.path], "y\n")
#         self.assertEqual(2, len(os.listdir(self.test_dir)))
#
#         # Test that we can build the resulting site
#         ret, _, _ = self.run_cli(["build"], cwd=self.path)
#         self.assertEqual(0, ret)
#
#     def test_custom_path(self):
#         """
#         Test that we can build a site in a custom folder.
#         """
#         ret, out, err = self.run_cli(["create", self.path, "--skeleton", os.path.join("cactus", "tests", "data", "skeleton")])
#         self.assertEqual(0, ret)
#
#         ret, out, err = self.run_cli(["build", "--path", self.path], cwd=self.test_dir)
#         self.assertEqual(0, ret)
#         self.assertIn(os.path.join(self.path, "config.json"), err)  # Check that we tried to find the config file in the right folder (and failed)
#
#         ret, _, _ = self.run_cli(["build", "-d", self.path], cwd=self.test_dir)
#         self.assertEqual(0, ret)
#
#         ret, _, _ = self.run_cli(["build"], cwd=self.test_dir)
#         self.assertNotEqual(0, ret)
#
#     def test_help(self):
#         ret, out, err = self.run_cli([])
#         self.assertNotEqual(0, ret)
#         self.assertIn("usage: cactus", err)
#
#     def test_serve(self):
#         cactus = self.find_cactus()
#
#         ret, _, _ = self.run_cli(["create", self.path])
#         self.assertEqual(0, ret)
#
#         port = 12345
#
#         p = subprocess.Popen([cactus, "serve", "-p", str(port)], cwd=self.path, stdout=subprocess.DEVNULL,
#                              stderr=subprocess.DEVNULL)
#
#         srv = "http://127.0.0.1:{0}".format(port)
#         s = requests.Session()
#         s.mount(srv, HTTPAdapter(max_retries=Retry(backoff_factor=0.2)))
#
#         r = s.post("{0}/_cactus/shutdown".format(srv))
#         r.raise_for_status()
#
#         # We'd love to use p.wait(n) here, but that doesn't work on
#         # some of the versions of Python we support.
#         for _ in range(5):
#             if p.poll() != None:
#                 break
#             time.sleep(1)
#         else:
#             self.fail("Server did not exit!")
#
#         self.assertEqual(0, p.returncode)



#coding:utf-8
from __future__ import unicode_literals

import os
import shutil
from six.moves import http_client, urllib

from cactus.site import Site
from cactus.plugin.manager import PluginManager
from cactus.utils.helpers import CaseInsensitiveDict
from cactus.utils.parallel import PARALLEL_DISABLED

from cactus.tests import BaseBootstrappedTestCase


class DummyPluginManager(PluginManager):
    """
    Doesn't do anything
    """
    def call(self, method, *args, **kwargs):
        """
        Trap the call
        """
        pass


class IntegrationTestCase(BaseBootstrappedTestCase):
    def setUp(self):
        super(IntegrationTestCase, self).setUp()

        self.site = Site(self.path,
            PluginManagerClass=DummyPluginManager, DeploymentEngineClass=self.get_deployment_engine_class())
        self.site._parallel = PARALLEL_DISABLED

        self.site.config.set('site-url', 'http://example.com/')

        # Clean up the site paths
        for path in (self.site.page_path, self.site.static_path):
            shutil.rmtree(path)
            os.mkdir(path)

    def get_deployment_engine_class(self):
        """
        Should return a deployment engine in tests.
        """
        pass


class BaseTestHTTPConnection(object):
    last_request = None

    def __init__(self, host, *args, **kwargs):
        self.host = host
        self.requests = []

    def connect(self):
        pass

    def close(self):
        pass

    def request(self, method, url, body=b'', headers=None):
        """
        Send a full request at once
        """
        if headers is None:
            headers = {}
        self.last_request = TestHTTPRequest(self, method, url, body, headers)

    def putrequest(self, method, url, *args, **kwargs):
        """
        Create a new request, but add more things to it later
        """
        self.current_request = TestHTTPRequest(self, method, url, b'', {})
        self.current_request.state = "headers"

    def putheader(self, header, value):
        """
        Add an header to a request that's in progress
        """
        self.current_request.headers[header] = value

    def endheaders(self, data=None):
        """
        End the headers of a request that's in progress
        """
        self.current_request.state = "body"
        self.last_request = self.current_request

        if data is not None:
            self.send(data)

    def send(self, data):
        """
        Add data to a request that's in progress
        """
        self.current_request.body += data

    def getresponse(self):
        request = self.last_request
        self.requests.append(request)
        return self.handle_request(request)

    def handle_request(self, request):
        """
        :param request: The request to handle
        """
        raise NotImplementedError("handle_request should be implemented by subclasses")


    def set_debuglevel(self, level):
        pass


class DebugHTTPSConnectionFactory(object):
    def __init__(self, conn_cls):
        self.conn_cls = conn_cls
        self.connections = []

    @property
    def requests(self):
        """
        :returns: A dictionary of the calls made through this connection factory (method -> list of calls)
        """
        out = []
        for connection in self.connections:
            out.extend(connection.requests)
        return out

    def __call__(self, *args, **kwargs):
        """
        Create a new connection from our connection class
        """
        connection = self.conn_cls(*args, **kwargs)
        self.connections.append(connection)
        return connection


class TestHTTPRequest(object):
    state = None

    def __init__(self, connection, method, url, body, headers):
        self.connection = connection

        self.method = method
        self.url = url
        self.body = body
        self.headers = CaseInsensitiveDict(headers)

        u = urllib.parse.urlparse(url)
        self.path = u.path
        self.params = urllib.parse.parse_qs(u.query, keep_blank_values=True)


class TestHTTPResponse(object):
    def __init__(self, status, reason=None, headers=None, body=''):
        if reason is None:
            reason = http_client.responses[status]
        if headers is None:
            headers = {}

        self.status = status
        self.reason = reason
        self.headers = CaseInsensitiveDict(headers)
        self.body = body

    def getheader(self, header, default=None):
        return self.headers.get(header, default)

    def getheaders(self):
        return self.headers

    def read(self):
        return self.body



import os
import pipes

os.environ['PATH'] = '/usr/bin:/bin:/usr/sbin:/sbin:/usr/local/bin:/usr/local/share/npm/bin:'

def postBuild(site):
    command = 'coffee -c %s/static/js/*.coffee' % pipes.quote(site.paths['build'])
    os.system(command)



import subprocess



KEY = 'GIT_COMMITID'

def commitid():
    command = 'git log -1 --format="%H"'
    print("run command [%s]" % command)
    result = subprocess.check_output(command, shell=True)
    result = result.rstrip('\n')
    print("result command [%s]" % result)
    return result


mycommitid = "undefined"

def preBuildPage(page, context, data):
    global mycommitid
    if (mycommitid == "undefined"):
        mycommitid = commitid()

    context[KEY] = mycommitid
    return context, data


#coding:utf-8
import os
import tempfile
import shutil
import unittest

import django.conf

from cactus.site import Site
from cactus.bootstrap import bootstrap
from cactus.config.router import ConfigRouter
from cactus.utils.parallel import PARALLEL_DISABLED


class BaseTestCase(unittest.TestCase):
    def setUp(self):
        self.test_dir = tempfile.mkdtemp()
        self.path = os.path.join(self.test_dir, 'test')
        self.clear_django_settings()

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def clear_django_settings(self):
        django.conf.settings._wrapped = django.conf.empty

    def assertFileExists(self, path):
        """
        Check that a file at path exists.
        """
        try:
            open(path)
        except IOError:
            path_dir = os.path.dirname(path)
            msg = [
                "File does not exist: {0}".format(path),
                "The following files *did* exist in {0}: {1}".format(path_dir, os.listdir(path_dir))
            ]
            self.fail("\n".join(msg))

    def assertFileDoesNotExist(self, path):
        """
        Check that the file at path does not exist.
        """
        try:
            open(path)
        except IOError:
            pass
        else:
            self.fail("File exists: {0}".format(path))


class BaseBootstrappedTestCase(BaseTestCase):
    def setUp(self):
        super(BaseBootstrappedTestCase, self).setUp()
        bootstrap(self.path, os.path.join("cactus", "tests", "data", "skeleton"))


class SiteTestCase(BaseBootstrappedTestCase):
    def setUp(self):
        super(SiteTestCase, self).setUp()
        self.config_path = os.path.join(self.path, 'config.json')
        self.conf = ConfigRouter([self.config_path])
        self.conf.set('site-url', 'http://example.com/')
        for k, v in self.get_config_for_test().items():
            self.conf.set(k, v)
        self.conf.write()

        self.site = Site(self.path, [self.config_path])
        self.site._parallel = PARALLEL_DISABLED

    def get_config_for_test(self):
        """
        Hook to set config keys in other tests.
        """
        return {}



#coding:utf-8
import logging
import time
from six.moves import urllib

from cactus.utils.parallel import multiMap


logger = logging.getLogger(__name__)


def retry(exceptions, tries=4, delay=3, backoff=2):
    """
    Retry execution in case we fail on one of the exceptions
    """
    def deco_retry(f):
        def f_retry(*args, **kwargs):
            mtries, mdelay = tries, delay
            try_one_last_time = True
            while mtries > 1:
                try:
                    return f(*args, **kwargs)
                except exceptions as e:
                    logger.warning("%s, Retrying in %.1f seconds..." % (str(e), mdelay))
                    time.sleep(mdelay)
                    mtries -= 1
                    mdelay *= backoff
            if try_one_last_time:
                return f(*args, **kwargs)
            return

        return f_retry  # true decorator

    return deco_retry


def internetWorking():
    def check(url):
        try:
            response = urllib.request.urlopen(url, timeout = 1)
            return True
        except urllib.error.URLError as err:
            pass
        return False

    return True in multiMap(check, [
        'http://www.google.com',
        'http://www.apple.com'])



#coding:utf-8
import os
import logging
import tempfile
import shutil

from cactus.compat.paths import StaticCompatibilityLayer
from cactus.utils.file import calculate_file_checksum, file_changed_hash
from cactus.utils.filesystem import alt_file, mkdtemp
from cactus.utils.url import ResourceURLHelperMixin


logger = logging.getLogger(__name__)

class Static(StaticCompatibilityLayer, ResourceURLHelperMixin):
    """
    A static resource in the repo
    """

    discarded = False

    def __init__(self, site, path, relative_to=None):
        """
        :param site: The site that's building this static file
        :param path: The location where this static file is to be found
        :param relative_to: Location this path is relative to. Optional, and defaults to the site's path.
        """
        self.site = site
        self.path = path
        self.relative_to = relative_to

        _static_path, filename = os.path.split(path)

        # Actual source file
        self.src_dir = os.path.join('static', _static_path)
        self.src_filename = filename

        try:
            self.src_name, self.src_extension = filename.rsplit('.', 1)
        except ValueError:
            self.src_name = filename
            self.src_extension = ""

        # Useless, we'll crash before.
        # # TODO
        # assert self.src_extension, "No extension for file?! {0}".format(self.src_name)

        # Do some pre-processing (e.g. optimizations):
        # must be done before fingerprinting
        self._preprocessing_path = self.pre_process()

        # Where the file will have to be referenced in output files

        if self.final_extension in self.site.fingerprint_extensions:
            checksum = calculate_file_checksum(self._preprocessing_path)
            new_name = "{0}.{1}".format(self.src_name, checksum)
        else:
            new_name = self.src_name

        # Path where this file should be referenced in source files
        self.link_url = '/' + os.path.join(self.src_dir, '{0}.{1}'.format(self.src_name, self.final_extension))

        self.final_name = "{0}.{1}".format(new_name, self.final_extension)

        if not hasattr(self.site, "_static_file_cache"):
            self.site._static_file_cache = {}


    @property
    def full_source_path(self):
        if self.relative_to is not None:
            relative_to = self.relative_to
        else:
            relative_to = self.site.path

        full_source_path = os.path.join(relative_to, self.src_dir, self.src_filename)

        if os.path.islink(full_source_path):
            return os.path.realpath(full_source_path)

        return full_source_path

    @property
    def build_path(self):
        """
        Path where the file should be built to.
        """
        return  os.path.join(self.src_dir, self.final_name)

    @property
    def full_build_path(self):
        """
        Absolute where the file should be built to.
        """
        return os.path.join(self.site.build_path, self.build_path)

    @property
    def _final_url(self):
        """
        Path where the file should be referenced in built files
        """
        return "/{0}".format(self.build_path)


    def run_externals(self, current_extension, pre_path, externals):
        """
        Run a set of externals against the file at pre_path
        Only one external will run (the first one to accept the file)

        Return the new extension for the file.
        """
        with alt_file(pre_path) as tmp_file:
            for ExternalClass in externals:
                external = ExternalClass(current_extension, pre_path, tmp_file)
                external.run()

                if external.accepted():
                    return external.output_extension
                elif external.refused():
                    continue
                elif external.discarded():
                    self.discard()
                    break

                raise Exception("External {0} has an unknown status: {1}".format(external, external.status))

            return current_extension

    def pre_process(self):
        """
        Does file pre-processing if required
        """
        self.pre_dir = mkdtemp()
        pre_path = os.path.join(self.pre_dir, self.src_filename)

        shutil.copy(self.full_source_path, pre_path)

        # Pre-process
        logger.debug('Pre-processing: %s %s', self.src_name, self.pre_dir)

        # Run processors (those might change the extension)
        self.final_extension = self.run_externals(self.src_extension, pre_path, self.site.external_manager.processors)

        # Optimize
        if not self.discarded:
            # Run optimizes and make sure they don't alter the extension
            _ = self.run_externals(self.final_extension, pre_path, self.site.external_manager.optimizers)
            assert self.final_extension == _, "Illegal Optimizer: may not change the extension"
            assert not self.discarded, "Illegal Optimizer: may not discard files"

        return pre_path

    def discard(self):
        self.discarded = True  #TODO: Warn on usage of the static!

    def build(self):

        # See if we can maybe skip this if the file did not change
        curr_hash = file_changed_hash(self.full_source_path)
        prev_hash = self.site._static_file_cache.get(self.full_source_path)

        if os.path.exists(self.full_build_path):
            if curr_hash == prev_hash:
                logger.debug("skip building (unchanged) %s %s", self.src_name, self.final_url)
                return

        self.site._static_file_cache[self.full_source_path] = curr_hash

        self.site.plugin_manager.preBuildStatic(self)

        if self.discarded:
            return

        logger.debug('Building {0} --> {1}'.format(self.src_name, self.full_build_path))

        try:
            os.makedirs(os.path.dirname(self.full_build_path))
        except OSError:
            pass

        shutil.copy(self._preprocessing_path, self.full_build_path)

        # self.site.plugin_manager.postBuildStatic(self)

    def __repr__(self):
        return '<Static: {0}>'.format(self.src_filename)



#coding:utf-8
import os

from cactus.tests.deployment import BaseDeploymentTestCase, DummyDeploymentEngine, DummySite, DummyUI, DummyFile


class FileChangedTestCase(BaseDeploymentTestCase):
    def setUp(self):
        super(FileChangedTestCase, self).setUp()
        self.ui = DummyUI()
        self.site = DummySite(self.test_dir, self.ui)
        self.engine = DummyDeploymentEngine(self.site)
        with open(os.path.join(self.test_dir, "123.html"), "w") as f:
            f.write("Hello!")

    def test_file_unchanged(self):
        """
        Test that we don't attempt to deploy unchanged files
        """
        class TestFile(DummyFile):
            def remote_changed(self):
                super(TestFile, self).remote_changed()
                return False

        self.engine.FileClass = TestFile
        self.engine.deploy()
        files = self.engine.created_files
        self.assertEqual(1, len(files))

        f = files[0]

        self.assertEqual(1, f.remote_changed_calls)
        self.assertEqual(0, f.do_upload_calls)


    def test_file_changed(self):
        """
        Test that we deploy files that changed
        """
        class TestFile(DummyFile):
            def remote_changed(self):
                super(TestFile, self).remote_changed()
                return True

        self.engine.FileClass = TestFile
        self.engine.deploy()
        files = self.engine.created_files
        self.assertEqual(1, len(files))

        f = files[0]

        self.assertEqual(1, f.remote_changed_calls)
        self.assertEqual(1, f.do_upload_calls)



# coding:utf-8
import os
import shutil

from cactus.tests import SiteTestCase
from cactus.utils.filesystem import fileList
from cactus.utils.url import path_to_url


class TestSite(SiteTestCase):
    def testBuild(self):
        """
        Test that we build the proper files.
        """
        self.site.build()

        # Make sure we build to .build and not build
        self.assertEqual(os.path.exists(os.path.join(self.path, 'build')), False)

        self.assertEqual(
            sorted([path_to_url(path) for path in fileList(os.path.join(self.path, '.build'), relative=True)]),
            sorted([
                'error.html',
                'index.html',
                'robots.txt',
                'sitemap.xml',
                self.site.get_url_for_static('/static/css/style.css')[1:],  # Strip the initial /
                self.site.get_url_for_static('/static/images/favicon.ico')[1:],  # Strip the initial /
                self.site.get_url_for_static('/static/js/main.js')[1:],  # Strip the initial /
            ]))

    def testRenderPage(self):
        """
        Test that pages get rendered.
        """

        shutil.copy(
            os.path.join('cactus', 'tests', 'data', "test-in.html"),
            os.path.join(self.path, 'pages', 'test.html')
        )

        self.site.build()

        with open(os.path.join('cactus', 'tests', 'data', 'test-out.html'), "rb") as expected:
            with open(os.path.join(self.path, '.build', 'test.html'), "rb") as obtained:
                self.assertEqual(expected.read(), obtained.read())

    def testPageContext(self):
        """
        Test that page context is parsed and uses in the pages.
        """

        shutil.copy(
            os.path.join('cactus', 'tests', 'data', "koenpage-in.html"),
            os.path.join(self.path, 'pages', 'koenpage.html')
        )

        self.site.build()

        with open(os.path.join('cactus', 'tests', 'data', 'koenpage-out.html'), "rb") as expected:
            with open(os.path.join(self.path, '.build', 'koenpage.html'), "rb") as obtained:
                self.assertEqual(expected.read(), obtained.read())

    def test_html_only_context(self):
        """
        Test that we only parse context on HTML pages.
        """
        robots_txt = 'Disallow:/page1\nDisallow:/page2'

        with open(os.path.join(self.site.path, 'pages', 'robots.txt'), 'w') as f:
            f.write(robots_txt)

        self.site.build()

        with open(os.path.join(self.site.build_path, 'robots.txt'), 'r') as f:
            self.assertEquals(robots_txt, f.read())

    def testStaticLoader(self):
        """
        Test that the static URL builde and template tag work.
        """
        static = '/static/css/style.css'
        page = "{%% static '%s' %%}" % static

        with open(os.path.join(self.path, 'pages', 'staticpage.html'), "w") as dst:
            dst.write(page)

        self.site.build()

        with open(os.path.join(self.path, '.build', 'staticpage.html'), "r") as obtained:
            self.assertEqual(self.site.get_url_for_static(static), obtained.read())

    def test_current_page(self):
        """
        Test that the if_current_page tag works.
        """
        page = 'page1.html'
        content = "{%% if_current_page '/%s' %%}" % page

        other = 'page2.html'

        with open(os.path.join(self.path, 'pages', page), 'w') as f:
            f.write(content)

        with open(os.path.join(self.path, 'pages', other), 'w') as f:
            f.write(content)

        self.site.build()

        with open(os.path.join(self.path, '.build', page), 'r') as f:
            self.assertEqual('True', f.read())

        with open(os.path.join(self.path, '.build', other), 'r') as f:
            self.assertEqual('False', f.read())



#coding:utf-8

# Define no-op plugin methods
def preBuildPage(page, context, data):
    """
    Called prior to building a page.

    :param page: The page about to be built
    :param context: The context for this page (you can modify this, but you must return it)
    :param data: The raw body for this page (you can modify this).
    :returns: Modified (or not) context and data.
    """
    return context, data


def postBuildPage(page):
    """
    Called after building a page.

    :param page: The page that was just built.
    :returns: None
    """
    pass


def preBuildStatic(static):
    """
    Called before building (copying to the build folder) a static file.

    :param static: The static file about to be built.
    :returns: None
    """
    pass


def postBuildStatic(static):
    """
    Called after building (copying to the build folder) a static file.

    :param static: The static file that was just built.
    :returns: None
    """
    pass


def preBuild(site):
    """
    Called prior to building the site, after loading configuration and plugins.

    A good time to register your externals.

    :param site: The site about to be built.
    :returns: None
    """
    pass

def postBuild(site):
    """
    Called after building the site.

    :param site: The site that was just built.
    :returns: None
    """
    pass


def preDeploy(site):
    """
    Called prior to deploying the site (built files)

    A good time to configure custom headers

    :param site: The site about to be deployed.
    :returns: None
    """
    pass


def postDeploy(site):
    """
    Called after deploying the site (built files)

    :param site: The site that was just built.
    :returns: None
    """
    pass


def preDeployFile(file):
    """
    Called prior to deploying a single built file

    :param file: The file about to be deployed.
    :returns: None
    """
    pass


ORDER = -1


DEFAULTS = [
    'preBuildPage',
    'postBuildPage',
    'preBuildStatic',
    'postBuildStatic',
    'preBuild',
    'postBuild',
    'preDeploy',
    'postDeploy',
    'preDeployFile',
]



#coding:utf-8
import os
import tempfile

from cactus.tests import SiteTestCase
from cactus.tests.compat import has_symlink
import unittest

class TestBuild(SiteTestCase):

    @unittest.skipUnless(has_symlink, "No symlink support")
    def test_existing_symlink(self):

        with open(os.path.join(self.site.static_path, 'file.js'), "w") as f:
            f.write("hello")

        os.symlink(
            os.path.join(self.site.static_path, 'file.js'),
            os.path.join(self.site.static_path, 'file-link.js'))

        self.site.build()

        self.assertFileExists(os.path.join(self.site.build_path, 'static', 'file.js'))
        self.assertFileExists(os.path.join(self.site.build_path, 'static', 'file-link.js'))

        with open(os.path.join(self.site.build_path, 'static', 'file-link.js')) as f:
            self.assertEqual(f.read(), "hello")

        self.assertEqual(
            os.path.islink(os.path.join(self.site.build_path, 'static', 'file-link.js')), False)

    @unittest.skipUnless(has_symlink, "No symlink support")
    def test_nonexisting_symlink(self):

        os.symlink(
            os.path.join(self.site.static_path, 'file.js'),
            os.path.join(self.site.static_path, 'file-link.js'))

        self.site.build()

        self.assertFileDoesNotExist(os.path.join(self.site.build_path, 'static', 'file.js'))
        self.assertFileDoesNotExist(os.path.join(self.site.build_path, 'static', 'file-link.js'))

    def test_pages_binary_file(self):

        with open(os.path.join(self.site.page_path, 'file.zip'), "wb") as f:
            f.write(os.urandom(1024))

        self.site.build()

    def test_listener_ignores(self):

        # Some complete;y random files
        self.assertEqual(True, self.site._rebuild_should_ignore("/Users/test/a.html"))
        self.assertEqual(True, self.site._rebuild_should_ignore("a.html"))
        self.assertEqual(True, self.site._rebuild_should_ignore("/a.html"))

        # Some plausible files
        self.assertEqual(True, self.site._rebuild_should_ignore(
            os.path.join(self.site.path, "config.json")))
        self.assertEqual(True, self.site._rebuild_should_ignore(
            os.path.join(self.site.path, "readme.txt")))
        self.assertEqual(True, self.site._rebuild_should_ignore(
            os.path.join(self.site.path, ".git", "config")))
        self.assertEqual(True, self.site._rebuild_should_ignore(
            os.path.join(self.site.path, ".build", "index.html")))
        self.assertEqual(True, self.site._rebuild_should_ignore(
            os.path.join(self.site.path, ".build", "static", "main.js")))

        # Files we do want
        self.assertEqual(False, self.site._rebuild_should_ignore(
            os.path.join(self.site.path, "pages", "index.html")))
        self.assertEqual(False, self.site._rebuild_should_ignore(
            os.path.join(self.site.path, "pages", "staging", "index.html")))
        self.assertEqual(False, self.site._rebuild_should_ignore(
            os.path.join(self.site.path, "templates", "base.html")))
        self.assertEqual(False, self.site._rebuild_should_ignore(
            os.path.join(self.site.path, "static", "main.js")))
        self.assertEqual(False, self.site._rebuild_should_ignore(
            os.path.join(self.site.path, "plugins", "test.py")))

        # Dotfile stuff
        self.assertEqual(True, self.site._rebuild_should_ignore(
            os.path.join(self.site.path, "pages", ".DS_Store")))
        self.assertEqual(True, self.site._rebuild_should_ignore(
            os.path.join(self.site.path, "pages", ".hidden", "a.html")))

    @unittest.skipUnless(has_symlink, "No symlink support")
    def test_symlink_ignore(self):

        file_path = os.path.join(tempfile.mkdtemp(), 'file.js')
        link_path = os.path.join(self.site.static_path, 'file-link.js')

        with open(file_path, "w") as f:
            f.write("hello")

        os.symlink(file_path, link_path)

        # Test for ignore function to work with symlinks
        self.assertEqual(self.site._rebuild_should_ignore(link_path), False)



#coding:utf-8
import os
import posixpath #TODO: Windows?
import logging

from cactus.utils.packaging import pkg_walk


logger = logging.getLogger(__name__)


def bootstrap_from_package(path):
    for dir_, sub_dirs, filenames in pkg_walk("cactus", "skeleton"):
        base_path = os.path.join(path, dir_.split('skeleton', 1)[1].lstrip('/'))
        for sub_dir in sub_dirs:
            dir_path = os.path.join(base_path, sub_dir)
            logger.debug("Creating {0}".format(dir_path))
            os.makedirs(dir_path)

        for filename in filenames:
            resource_path = posixpath.join(dir_, filename)
            file_path = os.path.join(base_path, filename)

            logger.debug("Copying {0} to {1}".format(resource_path, file_path))
            with open(file_path, 'wb') as f_writer:
                with open(os.path.join("cactus", resource_path), "rb") as f_reader:
                    f_writer.write(f_reader.read())



#coding:utf-8
from __future__ import print_function
import logging

from boto import s3
from boto.exception import S3ResponseError, S3CreateError
from boto.route53.exception import DNSServerError

from cactus.deployment.engine import BaseDeploymentEngine
from cactus.deployment.s3.auth import AWSCredentialsManager
from cactus.deployment.s3.file import S3File
from cactus.deployment.s3.domain import AWSBucket, AWSDomain
from cactus.exceptions import InvalidCredentials
from cactus.utils import ipc


logger = logging.getLogger(__name__)


class S3DeploymentEngine(BaseDeploymentEngine):
    FileClass = S3File
    CredentialsManagerClass = AWSCredentialsManager

    config_bucket_name = "aws-bucket-name"
    config_bucket_website = "aws-bucket-website"
    config_bucket_region = "aws-bucket-region"

    _s3_default_region = "us-east-1"
    _s3_port = 443
    _s3_is_secure = True
    _s3_https_connection_factory = None


    def _get_buckets(self):
        """
        :returns: The list of buckets found for this account
        """
        try:
            return self.get_connection().get_all_buckets()
        except S3ResponseError as e:
            if e.error_code == u'InvalidAccessKeyId':
                logger.info("Received an Error from AWS:\n %s", e.body)
                raise InvalidCredentials()
            raise

    def _get_bucket_region(self):
        return self.site.config.get(self.config_bucket_region, self._s3_default_region)

    def _create_connection(self):
        """
        Create a new S3 Connection
        """
        aws_access_key, aws_secret_key = self.credentials_manager.get_credentials()

        return s3.connect_to_region(self._get_bucket_region(),
                aws_access_key_id=aws_access_key.strip(),
                aws_secret_access_key=aws_secret_key.strip(),
                is_secure=self._s3_is_secure, port=self._s3_port,
                https_connection_factory=self._s3_https_connection_factory,
                calling_format='boto.s3.connection.OrdinaryCallingFormat'
        )

    def get_bucket(self):
        """
        :returns: The Bucket if found, None otherwise.
        :raises: InvalidCredentials if we can't connect to AWS
        """
        buckets = self._get_buckets()
        buckets = dict((bucket.name, bucket) for bucket in buckets)
        return buckets.get(self.bucket_name)

    def create_bucket(self):
        """
        :returns: The newly created bucket
        """
        try:

            # When creating a bucket, the region cannot be "us-east-1" but needs
            # to be an empty string, so we do that for now.
            # https://github.com/boto/boto3/issues/125#issuecomment-109408790
            if self._get_bucket_region() == "us-east-1":
                region = ""
            else:
                region = self._get_bucket_region()

            bucket = self.get_connection().create_bucket(self.bucket_name,
                policy='public-read', location=region
            )
        except S3CreateError:
            logger.info(
                    'Bucket with name %s already is used by someone else, '
                    'please try again with another name', self.bucket_name)
            return  #TODO: These should be exceptions

        # Configure S3 to use the index.html and error.html files for indexes and 404/500s.
        bucket.configure_website(self._index_page, self._error_page)

        return bucket

    def get_website_endpoint(self):
        return self.bucket.get_website_endpoint()

    def domain_setup(self):
        bucket_name = self.site.config.get(self.config_bucket_name)
        bucket_name = self.site.config.get(self.config_bucket_name)

        if not bucket_name:
            logger.warning("No bucket name")
            return

        aws_access_key, aws_secret_key = self.credentials_manager.get_credentials()
        domain = AWSDomain(aws_access_key, aws_secret_key, bucket_name)

        try:
            domain.setup()
        except DNSServerError as e:
            logger.debug(e)
            ipc.signal("domain.setup.error", {"errorKey": "AccountDisabled"})
            logger.error("Account cannot use route 53")
            logger.error(e)

    def domain_list(self):
        bucket_name = self.site.config.get(self.config_bucket_name)

        if not bucket_name:
            logger.warning("No bucket name")
            return

        aws_access_key, aws_secret_key = self.credentials_manager.get_credentials()
        domain = AWSDomain(aws_access_key, aws_secret_key, bucket_name)

        try:
            domain_list = domain.nameServers()
        except DNSServerError as e:
            print(e)
            ipc.signal("domain.list.error", {"errorKey": "AccountDisabled"})
            logger.error("Account cannot use route 53")
            logger.error(e)
            return

        if domain_list:
            ipc.signal("domain.list.result", {"nameservers": domain_list})
            for domain in domain_list:
                logger.info(domain)
        else:
            logger.error("No name servers configured")


    def domain_remove(self):
        pass



#coding:utf-8

class ContextPlugin(object):
    """
    A plugin to manage custom context via config files.

    The context can be made available via a "context" key in config files.
    """

    def preBuild(self, site):
        """
        Load the context from the config
        """
        self.context = site.config.get("context", {}, nested=True)

    def preBuildPage(self, page, context, data):
        """
        Update the page context with the config context
        """
        context.update(self.context)

        return context, data


plugin = ContextPlugin()



#coding:utf-8
import fnmatch

class IgnorePatternsPlugin(object):
    """
    Define configurable ignore patterns for static files and pages
    """
    def preBuild(self, site):
        """
        Load the ignore patterns from the site config
        """
        self.ignore_patterns = site.config.get('ignore', [])

    def preBuildPage(self, page, context, data):
        if not self.accept_path(page.source_path):
            page.discarded = True
        return context, data


    def preBuildStatic(self, static):
        if not self.accept_path(static.path):
            static.discard()

    def accept_path(self, path):
        """
        :param path: A path to be tested
        :returns: Whether this path can be includes in the build
        """
        for pattern in self.ignore_patterns:
            if fnmatch.fnmatch(path, pattern):
                return False
        return True



#coding:utf-8
import os
import shutil
import tempfile
import unittest

from cactus.config.file import ConfigFile
from cactus.config.router import ConfigRouter


class TestConfigRouter(unittest.TestCase):
    """
    Test that the config router manages multiple files correctly.
    """
    def setUp(self):
        self.test_dir = tempfile.mkdtemp()
        self.path = os.path.join(self.test_dir, "test")
        os.mkdir(self.path)

        self.path1 = os.path.join(self.path, "conf1.json")
        self.path2 = os.path.join(self.path, "conf2.json")
        self.conf1 = ConfigFile(self.path1)
        self.conf2 = ConfigFile(self.path2)

        self.conf1.set("a", 1)
        self.conf1.write()
        self.conf2.set("b", 2)
        self.conf2.write()

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def test_read(self):
        """
        Check that the config router reads correctly from the filesystem
        """
        router = ConfigRouter([self.path1, self.path2])

        self.assertEqual(router.get("a"), 1)
        self.assertEqual(router.get("b"), 2)
        self.assertEqual(router.get("c"), None)

    def test_read_write(self):
        """
        Check that our config is readable after writing it
        """
        router = ConfigRouter([self.path1, self.path2])

        router.set("a", 3)
        router.set("b", 4)

        self.assertEqual(3, router.get("a"))
        self.assertEqual(4, router.get("b"))

    def test_write(self):
        """
        Check that the config router writes correctly to the filesystem
        """
        router = ConfigRouter([self.path1, self.path2])
        router.set("a", 3)
        router.set("b", 4)
        router.write()

        self.conf1.load()
        self.conf2.load()

        self.assertEqual(self.conf1.get("a"), 3)
        self.assertEqual(self.conf1.get("b"), None)
        self.assertEqual(self.conf2.get("b"), 4)
        self.assertEqual(self.conf2.get("a"), None)

    def test_collision(self):
        """
        Check that we get the right key when there is a collision
        """
        self.conf1.set("b", 3)
        self.conf2.set("a", 4)
        self.conf1.write()
        self.conf2.write()

        router = ConfigRouter([self.path1, self.path2])

        self.assertEqual(router.get("a"), 1)
        self.assertEqual(router.get("b"), 3)

    def test_duplicate(self):
        """
        Check that the config router handles duplicate files properly.
        """
        router = ConfigRouter([self.path1, self.path1])
        router.set("a", 3)
        router.write()

        self.conf1.load()
        self.assertEqual(self.conf1.get("a"), 3)

    def test_nested(self):
        """
        Test that we support nested config for context
        """
        self.conf1.set("context", {"k1":"v1"})
        self.conf2.set("context", {"k2":"v2"})
        self.conf1.write()
        self.conf2.write()

        router = ConfigRouter([self.path1, self.path2])
        context = router.get("context", default={}, nested=True)

        self.assertEqual(context.get("k1"), "v1")
        self.assertEqual(context.get("k2"), "v2")

    def test_dirty(self):
        """
        Test that we don't re-write files that we haven't changed
        """

        self.conf1.set("a", "b")
        self.conf1.write()

        with open(self.path1, "w") as f:
            f.write("canary")

        self.conf1.write()

        with open(self.path1) as f:
            self.assertEqual("canary", f.read())

    def test_missing_file(self):
        """
        Test that we don't throw on a missing file, and that the configuration
        remains in a consistent state.
        """
        wrong_path = os.path.join(self.path, "does_not_exist.json")

        self.conf1.set("context", {"k1":"v1"})
        self.conf1.write()

        router = ConfigRouter([wrong_path, self.path1])

        self.assertEqual(router.get("context").get("k1"), "v1")

    def test_broken_file(self):
        """
        Test that we don't throw on a broken file, and that the configuration
        remains in a consistent state.
        """

        with open(self.path1, "w") as f:
            f.write("{broken}")

        self.conf2.set("context", {"k1":"v1"})
        self.conf2.write()

        router = ConfigRouter([self.path1, self.path2])

        self.assertEqual(router.get("context").get("k1"), "v1")




#coding:utf-8
import logging

from cactus.bootstrap.archive import bootstrap_from_archive
from cactus.bootstrap.package import bootstrap_from_package


logger = logging.getLogger(__name__)


def bootstrap(path, skeleton=None):
    """
    Bootstrap a new project at a given path.

    :param path: The location where the new project should be created.
    :param skeleton: An optional path to an archive that should be used instead of the standard cactus skeleton.
    """

    if skeleton is None:
        bootstrap_from_package(path)
    else:
        bootstrap_from_archive(path, skeleton)

    logger.info('New project generated at %s', path)



import logging
from cactus.static.external.exceptions import ExternalFailure


logger = logging.getLogger(__name__)


# Helpers to build the Externals

ACCEPTED = 1
REFUSED = 0
DISCARDED = -1


def status_getter(status):
    def has_status(external):
        return external.status == status
    return has_status


def status_setter(status):
    def set_status(external):
        logger.debug('External {0} ({1}) status > {2}'.format(
            external.__class__.__name__, external.src, status))
        external.status = status
    return set_status


class External(object):
    supported_extensions = ()  # The extensions supported by this output
    output_extension = 'css'  # The extension of this processor's output
    critical = False  # Whether this External failure is critical

    def __init__(self, extension, src, dst):
        self.extension = extension
        self.src = src
        self.dst = dst

    accept = status_setter(ACCEPTED)
    accepted = status_getter(ACCEPTED)

    refuse = status_setter(REFUSED)
    refused = status_getter(REFUSED)

    discard = status_setter(DISCARDED)
    discarded = status_getter(DISCARDED)


    def run(self):
        """
        Return True in the case we succeed in running, False otherwise.
        This means we can use several processors and have one or the other work.
        """
        if not self.extension in self.supported_extensions:
            return self.refuse()

        self.accept()  # We accept now so the run method can discard

        try:
            self._run()
        except OSError as e:
            msg = 'Could not call external processor {0}: {1}'.format(self.__class__.__name__, e)

            if self.critical:
                logger.critical(msg)
                raise ExternalFailure(self.__class__.__name__, e)
            else:
                logger.info(msg)
                self.refuse()

    def _run(self):
        raise NotImplementedError()



# Python 3.5 Django compatibility fix:
# + https://github.com/django/django/commit/b07aa52e8a8e4c7fdc7265f75ce2e7992e657ae9)
# + https://code.djangoproject.com/ticket/23763
import six

if six.PY3:
    import html.parser as _html_parser

    try:
        HTMLParseError = _html_parser.HTMLParseError
    except AttributeError:
        # create a dummy class for Python 3.5+ where it's been removed
        class HTMLParseError(Exception):
            pass

        _html_parser.HTMLParseError = HTMLParseError




#coding:utf-8
from cactus.tests.deployment import DummyUI, DummySite, DummyDeploymentEngine, BaseDeploymentTestCase


class BucketNameTestCase(BaseDeploymentTestCase):
    def setUp(self):
        super(BucketNameTestCase, self).setUp()
        self.ui = DummyUI(create_bucket=False)
        self.site = DummySite(self.test_dir, self.ui)
        self.engine = DummyDeploymentEngine(self.site)

    def test_not_configured(self):
        """
        Test that we prompt the bucket name in case it's not configured
        """
        self.assertEqual(0, self.ui.asked_name)

        self.engine.configure()

        self.assertEqual(1, self.ui.asked_name)
        self.assertEqual("test-bucket", self.engine.bucket_name)

    def test_configured(self):
        """
        Test that we don't prompt the bucket name in case it's configured
        """
        self.site.config.set("test-conf-entry", "test-bucket")

        self.assertEqual(0, self.ui.asked_name)

        self.engine.configure()

        self.assertEqual(0, self.ui.asked_name)
        self.assertEqual("test-bucket", self.engine.bucket_name)



#coding:utf-8
from __future__ import unicode_literals

import os
import shutil
import tempfile
import hashlib
import unittest
import mock

from cactus.deployment.engine import BaseDeploymentEngine
from cactus.config.router import ConfigRouter
from cactus.deployment.file import BaseFile
from cactus.plugin.builtin.cache import CacheDurationPlugin
from cactus.plugin.loader import ObjectsPluginLoader
from cactus.plugin.manager import PluginManager


class TestHeadersPlugin(object):
    """
    An utility plugin to retrieve a file's header.
    """
    def __init__(self):
        self.headers = None

    def preDeployFile(self, file):
        self.file = file


class TestFile(BaseFile):
    def remote_changed(self):
        return True

    def do_upload(self):
        pass

class TestDeploymentEngine(BaseDeploymentEngine):
    CredentialsManagerClass = lambda self, engine: None  # We never use it here


#TODO: Retest this with a custom deployment engine or file class

class TestDeployFile(unittest.TestCase):
    def setUp(self):
        self.test_dir = tempfile.mkdtemp()
        self.build_path = os.path.join(self.test_dir, '.build')
        os.mkdir(self.build_path)


        self.site = mock.MagicMock()
        self.site.plugin_manager = PluginManager(self.site, [])
        self.site.path = self.test_dir
        self.site.build_path = self.build_path
        self.site.config = ConfigRouter([os.path.join(self.test_dir, "config.json")])
        self.site.config.set("site-url", "http://example.com")

        self.engine = TestDeploymentEngine(self.site)


    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def test_cache_control(self):
        """
        Ensure that the Cache control headers are properly set
        """
        # Test a fingerprinted file
        content = 'abc'
        h = hashlib.md5(content.encode('utf-8')).hexdigest()
        filename = "file.{0}.data".format(h)

        with open(os.path.join(self.build_path,  filename), "w") as f:
            f.write(content)

        p = TestHeadersPlugin()
        self.site.plugin_manager.loaders = [ObjectsPluginLoader([p, CacheDurationPlugin()])]
        self.site.plugin_manager.reload()
        self.site.plugin_manager.preDeploy(self.site)

        f = TestFile(self.engine, filename)
        f.upload()

        self.assertEqual(p.file.cache_control, f.MAX_CACHE_EXPIRATION)


        # Test a non fingerprinted file
        with open(os.path.join(self.build_path, "123.html"), "w") as f:
            f.write("abc")

        # Prepare setup
        p = TestHeadersPlugin()
        self.site.plugin_manager.loaders = [ObjectsPluginLoader([p, CacheDurationPlugin()])]
        self.site.plugin_manager.reload()
        f = TestFile(self.engine, "123.html")

        # Test with no configured cache duration
        self.site.config.set("cache-duration", None)
        self.site.plugin_manager.preDeploy(self.site)

        f.upload()
        self.assertEqual(p.file.cache_control, f.DEFAULT_CACHE_EXPIRATION)

        # Test with a configured cache duration
        self.site.config.set("cache-duration", 123)
        self.site.plugin_manager.preDeploy(self.site)

        f.upload()
        self.assertEqual(p.file.cache_control, 123)



import os
import io
import logging

import yaml

from six.moves import urllib

from django.template import Template, Context
from cactus.compat.paths import PageCompatibilityLayer
from cactus.utils.url import ResourceURLHelperMixin


logger = logging.getLogger(__name__)


class Page(PageCompatibilityLayer, ResourceURLHelperMixin):

    discarded = False

    def __init__(self, site, source_path):
        self.site = site

        # The path where this element should be linked in "base" pages
        self.source_path = source_path

        # The URL where this element should be linked in "base" pages
        self.link_url = '/{0}'.format(self.source_path)

        if self.site.prettify_urls:
            # The URL where this element should be linked in "built" pages
            if self.is_html():
                if self.is_index():
                    self.final_url = self.link_url.rsplit('index.html', 1)[0]
                else:
                    self.final_url = '{0}/'.format(self.link_url.rsplit('.html', 1)[0])
            else:
                self.final_url = self.link_url

            # The path where this element should be built to
            if not self.is_html() or self.source_path.endswith('index.html'):
                self.build_path = self.source_path
            else:
                self.build_path = '{0}/{1}'.format(self.source_path.rsplit('.html', 1)[0], 'index.html')
        else:
            self.final_url = self.link_url
            self.build_path = self.source_path

    def is_html(self):
        return urllib.parse.urlparse(self.source_path).path.endswith('.html')

    def is_index(self):
        return urllib.parse.urlparse(self.source_path).path.endswith('index.html')

    @property
    def absolute_final_url(self):
        """
        Return the absolute URL for this page in the final build
        """
        return urllib.parse.urljoin(self.site.url, self.final_url)

    @property
    def full_source_path(self):
        return os.path.join(self.site.path, 'pages', self.source_path)

    @property
    def full_build_path(self):
        return os.path.join(self.site.build_path, self.build_path)

    def data(self):
        with io.FileIO(self.full_source_path, 'r') as f:
            try:
                return f.read().decode('utf-8')
            except:
                logger.warning("Template engine could not process page: %s", self.path, exc_info=True)
                return u""

    def context(self, data=None, extra=None):
        """
        The page context.
        """
        if extra is None:
            extra = {}

        context = {'__CACTUS_CURRENT_PAGE__': self,}

        page_context, data = self.parse_context(data or self.data())

        context.update(self.site.context())
        context.update(extra)
        context.update(page_context)

        return Context(context)

    def render(self):
        """
        Takes the template data with context and renders it to the final output file.
        """

        data = self.data()
        context = self.context(data=data)

        # This is not very nice, but we already used the header context in the
        # page context, so we don't need it anymore.
        page_context, data = self.parse_context(data)

        context, data = self.site.plugin_manager.preBuildPage(
            self.site, self, context, data)
        return Template(data).render(context)

    def build(self):
        """
        Save the rendered output to the output file.
        """
        logger.debug('Building {0} --> {1}'.format(self.source_path, self.final_url))  #TODO: Fix inconsistency w/ static
        data = self.render()  #TODO: This calls preBuild indirectly. Not great.

        if not self.discarded:

            # Make sure a folder for the output path exists
            try:
                os.makedirs(os.path.dirname(self.full_build_path))
            except OSError:
                pass

            with io.FileIO(self.full_build_path, 'w') as f:
                f.write(data.encode('utf-8'))

            self.site.plugin_manager.postBuildPage(self)

    def parse_context(self, data, splitChar=':'):
        """
        Values like

        name: koen
        age: 29

        will be converted in a dict: {'name': 'koen', 'age': '29'}
        Also: Blocks between --- and ... will be converted into a YAML objects
        """

        if not self.is_html():
            return {}, data

        lines = data.splitlines()
        if not lines:
            return {}, ''

        values = {}
        parsing_yaml = False
        for i, line in enumerate(lines):

            # YAML Document Start. Shortcircuit and parse as YAML instead
            if "---" in line:
                parsing_yaml = True
                break

            # Start Compatibility Parser
            elif not line:
                continue

            elif splitChar in line:
                line = line.split(splitChar)
                values[line[0].strip()] = (splitChar.join(line[1:])).strip()

            else:
                break

        # Parse as YAML
        if parsing_yaml:
            [raw_yaml, template] = data.split('...', 1)
            try:
                return yaml.load(raw_yaml), template
            except Exception as e:
                # Safety return
                return {}, template

        # Parse using the old parser
        else:
            return values, '\n'.join(lines[i:])

    def __repr__(self):
        return '<Page: {0}>'.format(self.source_path)



#coding:utf-8
import os
from cactus.tests import SiteTestCase


class TestIgnore(SiteTestCase):
    def test_ignore_static(self):
        with open(os.path.join(self.site.static_path, "koen.psd"), "w") as f:
            f.write("Not really a psd")

        with open(os.path.join(self.site.static_path, "koen.gif"), "w") as f:
            f.write("Not really a gif")

        self.site.config.set("ignore", ["*.psd"])
        self.site.build()

        self.assertFileDoesNotExist(os.path.join(self.site.build_path, "static", "koen.psd"))
        self.assertFileExists(os.path.join(self.site.build_path, "static", "koen.gif"))

    def test_ignore_pages(self):
        with open(os.path.join(self.site.page_path, "page.html.swp"), "w") as f:
            f.write("Not really a swap file")

        with open(os.path.join(self.site.page_path, "page.txt"), "w") as f:
            f.write("Actually a text file")

        self.site.config.set("ignore", ["*.swp"])
        self.site.build()

        self.assertFileDoesNotExist(os.path.join(self.site.build_path, "page.html.swp"))
        self.assertFileExists(os.path.join(self.site.build_path, "page.txt"))



#coding:utf-8
import subprocess
from cactus.static.external import External


class YUIJSOptimizer(External):
    supported_extensions = ('js',)
    output_extension = 'js'

    def _run(self):
        subprocess.call([
            'yuicompressor',
            '--type', 'js',
            '-o', self.dst,
            self.src,
        ])


class YUICSSOptimizer(External):
    supported_extensions = ('css',)
    output_extension = 'css'

    def _run(self):
        subprocess.call([
            'yuicompressor',
            '--type', 'css',
            '-o', self.dst,
            self.src,
        ])



import os
import logging
import types
import json

import six


class JsonFormatter(logging.Formatter):

    def format(self, record):

        data = {
            "level": record.levelno,
            "levelName": record.levelname,
            "msg": logging.Formatter.format(self, record)
        }

        if type(record.args) is types.DictType:
            for k, v in six.iteritems(record.args):
                data[k] = v

        return json.dumps(data)


def setup_logging(verbose, quiet):

    logger = logging.getLogger()
    handler = logging.StreamHandler()

    if os.environ.get('DESKTOPAPP'):
        log_level = logging.INFO
        handler.setFormatter(JsonFormatter())

    else:
        from colorlog import ColoredFormatter

        formatter = ColoredFormatter(
                "%(log_color)s%(message)s",
                datefmt=None,
                reset=True,
                log_colors={
                    'DEBUG':    'cyan',
                    'INFO':     'green',
                    'WARNING':  'yellow',
                    'ERROR':    'red',
                    'CRITICAL': 'bold_red',
                    }
                )

        if quiet:
            log_level = logging.WARNING
        elif verbose:
            log_level = logging.DEBUG
        else:
            log_level = logging.INFO

        handler.setFormatter(formatter)

    logger.setLevel(log_level)

    for h in logger.handlers:
        logger.removeHandler(h)

    logger.addHandler(handler)



#coding:utf-8
import logging


logger = logging.getLogger(__name__)


PROVIDER_MAPPING = {
    "rackspace": "cactus.deployment.cloudfiles.engine.CloudFilesDeploymentEngine",
    "google": "cactus.deployment.gcs.engine.GCSDeploymentEngine",
    "aws": "cactus.deployment.s3.engine.S3DeploymentEngine",
}


def get_deployment_engine_class(provider):
    """
    Import an engine by name
    :provider: The provider you want to deploy to
    :type provider: str
    :returns: The deployment Engine
    :rtype: cactus.deployment.engine.BaseDeploymentEngine
    """
    engine_path = PROVIDER_MAPPING.get(provider, None)
    logger.debug("Loading Deployment Engine for %s: %s", provider, engine_path)

    if engine_path is None:
        return None

    module, engine = engine_path.rsplit(".", 1)

    try:
        _mod = __import__(module, fromlist=[engine])
    except ImportError as e:
        logger.error("Unable to import requested engine (%s) for provider %s", engine, provider)
        logger.error("A required library was missing: %s", e.message)
        logger.error("Please install the library and try again")
    else:
        return getattr(_mod, engine)



import os
import pipes
import subprocess
import logging

def run(command):

    logger = logging.getLogger(__name__)

    logger.debug(command)

    os.environ['PATH'] = '/usr/bin:/bin:/usr/sbin:/sbin:/usr/local/bin:/usr/local/share/npm/bin:'

    process = subprocess.Popen([command],
        shell=True,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE)

    stdout = process.stdout.readline()
    stderr = process.stderr.readline()

    if stdout: logger.info(stdout)
    if stderr: logger.warning(stderr)


def preBuild(site):
    run('coffee -c %s/js/*.coffee' % pipes.quote(site.static_path))



import os
import sys
import shutil
import logging
import traceback
import django.conf
from django.utils import translation

from cactus import ui as ui_module
from cactus.config.router import ConfigRouter
from cactus.deployment import get_deployment_engine_class
from cactus.i18n.commands import MessageMaker, MessageCompiler
from cactus.plugin.builtin.cache import CacheDurationPlugin
from cactus.plugin.builtin.context import ContextPlugin
from cactus.plugin.builtin.ignore import IgnorePatternsPlugin
from cactus.plugin.loader import CustomPluginsLoader, ObjectsPluginLoader
from cactus.plugin.manager import PluginManager
from cactus.static.external.manager import ExternalManager
from cactus.compat.paths import SiteCompatibilityLayer
from cactus.compat.page import PageContextCompatibilityPlugin
from cactus.utils.file import fileSize
from cactus.utils.filesystem import chdir, fileList
from cactus.utils.helpers import memoize, map_apply
from cactus.utils.network import internetWorking
from cactus.utils.parallel import multiMap, PARALLEL_DISABLED, PARALLEL_CONSERVATIVE, PARALLEL_AGGRESSIVE
from cactus.utils.url import is_external
from cactus.page import Page
from cactus.static import Static
from cactus.listener import Listener
from cactus.server import WebServer
from cactus.utils import ipc


logger = logging.getLogger(__name__)

DEFAULT_PROVIDER = "aws"


class Site(SiteCompatibilityLayer):
    _path = None
    _parallel = PARALLEL_CONSERVATIVE  #TODO: Test me
    _static = None

    VERB_UNKNOWN = 0
    VERB_SERVE = 1
    VERB_BUILD = 2

    def __init__(self, path, config_paths=None, ui=None,
        PluginManagerClass=None, ExternalManagerClass=None, DeploymentEngineClass=None,
        verb=VERB_UNKNOWN):

        # Load the config engine
        if config_paths is None:
            config_paths = []
        self.config = ConfigRouter(config_paths)
        self.verb = verb

        # Load site-specific config values
        self.prettify_urls = self.config.get('prettify', False)
        self.compress_extensions = self.config.get('compress', ['html', 'css', 'js', 'txt', 'xml'])
        self.fingerprint_extensions = self.config.get('fingerprint', [])
        self.use_translate = self.config.get('use_translate', False)

        self.locale = []
        self.default_language = self.config.get('default_language', 'en')
        if self.use_translate:
            self.locale.append(self.default_language)
            self.other_languages = self.config.get('other_languages', [])
            self.locale += self.other_languages

        self.path = path
        self.verify_path()

        # Load Managers
        if ui is None:
            ui = ui_module
        self.ui = ui

        if PluginManagerClass is None:
            PluginManagerClass =  PluginManager
        self.plugin_manager = PluginManagerClass(self,
            [
                CustomPluginsLoader(self.plugin_path),  # User plugins
                ObjectsPluginLoader([   # Builtin plugins
                    ContextPlugin(), CacheDurationPlugin(),
                    IgnorePatternsPlugin(), PageContextCompatibilityPlugin(),
                ])
            ]
        )

        if ExternalManagerClass is None:
            ExternalManagerClass = ExternalManager
        self.external_manager = ExternalManagerClass(self)

        if DeploymentEngineClass is None:
            hosting_provider = self.config.get("provider", DEFAULT_PROVIDER)
            DeploymentEngineClass = get_deployment_engine_class(hosting_provider)
            assert DeploymentEngineClass is not None, \
                   "Could not load Deployment for Provider: {0}".format(hosting_provider)
        self.deployment_engine = DeploymentEngineClass(self)

        # Load Django settings
        self.setup()

    @property
    def url(self):
        return self.config.get('site-url')

    @url.setter
    def url(self, value):
        self.config.set('site-url', value)
        self.config.write()

    def verify_url(self):
        """
        We need the site url to generate the sitemap.
        """
        #TODO: Make a "required" option in the config.
        #TODO: Use URL tags in the sitemap

        # if self.url is None:
        #     self.url = self.ui.prompt_url("Enter your site URL (e.g. http://example.com/)")

    @property
    def path(self):
        return self._path

    @path.setter
    def path(self, path):
        self._path = path

        self.build_path = os.path.join(path, '.build')
        self.deploy_path = os.path.join(path, '.deploy')
        self.template_path = os.path.join(path, 'templates')
        self.page_path = os.path.join(path, 'pages')
        self.plugin_path = os.path.join(path, 'plugins')
        self.static_path = os.path.join(path, 'static')
        self.script_path = os.path.join(os.getcwd(), __file__)
        self.locale_path = os.path.join(path, "locale")


    def setup(self):
        """
        Configure django to use both our template and pages folder as locations
        to look for included templates.
        """
        TEMPLATES = [{
            'BACKEND': 'django.template.backends.django.DjangoTemplates',
            'DIRS': [self.template_path, self.page_path],
            'APP_DIRS': True,
            'OPTIONS': {
                'context_processors': [
                    'django.template.context_processors.debug',
                    'django.template.context_processors.request',
                    'django.contrib.auth.context_processors.auth',
                    'django.contrib.messages.context_processors.messages',
                    'myproject.utils.context_processors.settings_context',
                ],
                'builtins': [
                    'cactus.template_tags',
                ],
            },
        }]

        MIDDLEWARE = [
            'django.middleware.locale.LocaleMiddleware',
        ]

        settings = {
            "TEMPLATES": TEMPLATES,
            "MIDDLEWARE": MIDDLEWARE,
        }

        if self.use_translate:

            settings.update({
                "USE_I18N": True,
                "USE_L10N": False,
                "LOCALE_PATHS": [self.locale_path],
            })

        django.conf.settings.configure(**settings)
        django.setup()

    def verify_path(self):
        """
        Check if this path looks like a Cactus website
        """
        required_subfolders = ['pages', 'static', 'templates', 'plugins']
        if self.use_translate:
            required_subfolders.append('locale')

        for p in required_subfolders:
            if not os.path.isdir(os.path.join(self.path, p)):
                logger.error('This does not look like a (complete) cactus project (missing "%s" subfolder)', p)
                sys.exit(1)

    @memoize
    def context(self):
        """
        Base context for the site: all the html pages.
        """
        ctx = {
            'CACTUS': {
                'pages':  [p for p in self.pages() if p.is_html()],
                'static': [p for p in self.static()]
            },
            '__CACTUS_SITE__': self,
        }

        # Also make lowercase work
        ctx['cactus'] = ctx['CACTUS']

        return ctx

    def make_messages(self):
        """
        Generate the .po files for the site.
        """
        if not self.use_translate:
            logger.error("You should set use_translate=true in configurations file. "
                         "You also can set default_language and other_languages to translate")
            return

        message_maker = MessageMaker(self)
        message_maker.execute()

    def compile_messages(self):
        """
        Remove pre-existing compiled language files, and re-compile.
        """

        message_compiler = MessageCompiler(self)
        message_compiler.execute()

    def clean(self):
        """
        Remove all build files.
        """
        logger.debug("*** CLEAN %s", self.path)

        if os.path.isdir(self.build_path):
            shutil.rmtree(self.build_path)

    def build_with_translation(self, locale_item=None):
        logger.debug("*** BUILD %s", self.path)

        language = self.default_language
        if locale_item is not None:
            locale_build_path = os.path.join(self.build_path, locale_item)
            self.build_path = locale_build_path
            language = locale_item

        django.conf.settings.LANGUAGE_CODE = language
        translation.activate(language)

        self.verify_url()

        # Reset the static content
        self._static = None
        self._static_resources_dict = None

        # TODO: Facility to reset the site, and reload config.
        # TODO: Currently, we can't build a site instance multiple times
        self.plugin_manager.reload()  # Reload in case we're running on the server # We're still loading twice!

        self.plugin_manager.preBuild(self)

        logger.debug('Plugins:    %s', ', '.join([p.plugin_name for p in self.plugin_manager.plugins]))
        logger.debug('Processors: %s', ', '.join([p.__name__ for p in self.external_manager.processors]))
        logger.debug('Optimizers: %s', ', '.join([p.__name__ for p in self.external_manager.optimizers]))

        # Make sure the build path exists
        if not os.path.exists(self.build_path):
            os.mkdir(self.build_path)

        # Copy the static files
        self.buildStatic()

        # Always clean out the pages

        build_static_path = os.path.join(self.build_path, "static")

        for path in os.listdir(self.build_path):
            path = os.path.join(self.build_path, path)
            if path != build_static_path:
                if os.path.isdir(path):
                    shutil.rmtree(path)
                else:
                    os.remove(path)

        # Render the pages to their output files
        mapper = multiMap if self._parallel >= PARALLEL_AGGRESSIVE else map_apply
        # mapper = map_apply
        mapper(lambda p: p.build(), self.pages())

        self.plugin_manager.postBuild(self)

        for static in self.static():
            if os.path.isdir(static.pre_dir):
                shutil.rmtree(static.pre_dir)

    def build(self):
        """
        Generate fresh site from templates.
        """
        self.build_with_translation()

        # Prepare translations
        if self.use_translate:
            build_path_tmp = self.build_path
            self.compile_messages()

            for locale_item in self.locale:
                self.build_with_translation(locale_item)
                self.build_path = build_path_tmp

    def static(self):
        """
        Retrieve a list of static files for the site
        """
        if self._static is None:

            self._static = []

            for path in fileList(self.static_path, relative=True):

                full_path = os.path.join(self.static_path, path)

                if os.path.islink(full_path):
                    if not os.path.exists(os.path.realpath(full_path)):
                        logger.warning("Skipping symlink that points to unexisting file:\n%s", full_path)
                        continue

                self._static.append(Static(self, path))

        return self._static

    def static_resources_dict(self):
        """
        Retrieve a dictionary mapping URL's to static files
        """
        if self._static_resources_dict is None:
            self._static_resources_dict = dict((resource.link_url, resource) for resource in self.static())

        return self._static_resources_dict

    def _get_resource(self, src_url, resources):

        if is_external(src_url):
            return src_url

        for split_char in ["#", "?"]:
            if split_char in src_url:
                src_url = src_url.split(split_char)[0]

        if src_url in resources:
            return resources[src_url].final_url

        return None


    def _get_url(self, src_url, resources):
        return self._get_resource(src_url, resources)

    def get_url_for_static(self, src_path):
        return self._get_url(src_path, self.static_resources_dict())

    def get_url_for_page(self, src_path):
        return self._get_url(src_path, dict((resource.link_url, resource) for resource in self.pages()))

    def buildStatic(self):
        """
        Build static files (pre-process, copy to static folder)
        """
        mapper = multiMap if self._parallel > PARALLEL_DISABLED else map_apply
        mapper(lambda s: s.build(), self.static())

    def pages(self):
        """
        List of pages.
        """

        if not hasattr(self, "_page_cache"):
            self._page_cache = {}

        pages = []

        for path in fileList(self.page_path, relative=True):

            if path.endswith("~"):
                continue

            if path not in self._page_cache:
                logger.debug("Found page: %s", path)
                self._page_cache[path] = Page(self, path)

            pages.append(self._page_cache[path])

        return pages

    def _rebuild_should_ignore(self, file_path):

        file_relative_path = os.path.relpath(file_path, self.path)

        # Ignore anything in a hidden folder like .git
        for path_part in file_relative_path.split(os.path.sep):
            if path_part.startswith("."):
                return True

        if file_path.startswith(self.page_path):
            return False

        if file_path.startswith(self.template_path):
            return False

        if file_path.startswith(self.static_path):
            return False

        if file_path.startswith(self.plugin_path):
            return False

        return True

    def _rebuild(self, changes):

        logger.debug("*** REBUILD %s", self.path)

        logger.info('*** Rebuilding (%s changed)' % self.path)

        # We will pause the listener while building so scripts that alter the output
        # like coffeescript and less don't trigger the listener again immediately.
        self.listener.pause()

        try:
            #TODO: Fix this.
            #TODO: The static files should handle collection of their static folder on their own
            #TODO: The static files should not run everything on __init__
            #TODO: Only rebuild static files that changed
            # We need to "clear out" the list of static first. Otherwise, processors will not run again
            # They run on __init__ to run before fingerprinting, and the "built" static files themselves,
            # which are in a temporary folder, have been deleted already!
            # self._static = None
            self.build()

        except Exception as e:
            logger.info('*** Error while building\n%s', e)
            traceback.print_exc(file=sys.stdout)

        changed_file_extension = set(map(lambda x: os.path.splitext(x)[1], changes["changed"]))
        reload_css_file_extenstions = set([".css", ".sass", ".scss", ".styl"])

        # When we have changes, we want to refresh the browser tabs with the updates.
        # Mostly we just refresh the browser except when there are just css changes,
        # then we reload the css in place.

        local_hosts = [
            "http://127.0.0.1:%s" % self._port,
            "http://localhost:%s" % self._port,
            "http://0.0.0.0:%s" % self._port
        ]

        if len(changes["added"]) == 0 and len(changes["deleted"]) == 0 and changed_file_extension.issubset(reload_css_file_extenstions):
            # browserReloadCSS(local_hosts)
            self.server.reloadCSS()
        else:
            # browserReload(local_hosts)
            self.server.reloadPage()

        self.listener.resume()

    def serve(self, browser=True, port=8000):
        """
        Start a http server and rebuild on changes.
        """
        self._parallel = PARALLEL_DISABLED
        self._port = port
        self.verb = self.VERB_SERVE

        self.clean()
        self.build()

        logger.info('Running webserver at http://127.0.0.1:%s for %s' % (port, self.build_path))
        ipc.signal("server.didstart")
        logger.info('Type control-c to exit')

        with chdir(self.build_path):
            self.listener = Listener(self.path, self._rebuild, ignore=self._rebuild_should_ignore)
            self.listener.run()

        self.server = WebServer(self.build_path, port=port)

        try:
            self.server.start()

            # if browser is True:
            #     webbrowser.open('http://127.0.0.1:%s' % port)

        except (KeyboardInterrupt, SystemExit):
            self.server.stop()
            logger.info("Bye")

    def upload(self):

        # Make sure we have internet
        if not internetWorking():
            logger.info('There does not seem to be internet here, check your connection')
            return

        logger.debug('Start upload')

        self.build_path = self.deploy_path

        self.clean()
        self.build()

        self.plugin_manager.preDeploy(self)

        totalFiles = self.deployment_engine.deploy()
        changedFiles = [r for r in totalFiles if r['changed']]

        self.plugin_manager.postDeploy(self)

        # Display done message and some statistics
        logger.info('\nDone\n')

        logger.info('%s total files with a size of %s' %
                     (len(totalFiles), fileSize(sum([r['size'] for r in totalFiles]))))
        logger.info('%s changed files with a size of %s' %
                     (len(changedFiles), fileSize(sum([r['size'] for r in changedFiles]))))

        logger.info('\nhttp://%s\n' % self.config.get('aws-bucket-website'))  #TODO: Fix


    def domain_setup(self):

        # Make sure we have internet
        if not internetWorking():
            logger.info('There does not seem to be internet here, check your connection')
            return

        self.deployment_engine.domain_setup()
        self.domain_list()

    def domain_list(self):
        self.deployment_engine.domain_list()



#coding:utf-8

CACTUS_CLIENT_ID = "985227088845.apps.googleusercontent.com"
CACTUS_CLIENT_SECRET = "6PzihpbiC33TW-GIagIRH0t_"
CACTUS_REQUIRED_SCOPE = "https://www.googleapis.com/auth/devstorage.full_control"
LOCAL_REDIRECT_URI = "urn:ietf:wg:oauth:2.0:oob"



#coding:utf-8

class InvalidExternal(Exception):
    """
    Raised when an External caused an illegal operation.
    """
    #TODO: Tests


class ExternalFailure(Exception):
    """
    Raised when an External failed to run
    """
    def __init__(self, external, error):
        self.external = external
        self.error = error

    def __str__(self):
        return '{0} failed: {1}'.format(self.external, self.error)



#coding:utf-8
from __future__ import unicode_literals

import os
import io
import gzip

from six import BytesIO

from cactus.deployment.file import BaseFile

from cactus.tests.integration.s3 import S3IntegrationTestCase


class DeployTestCase(S3IntegrationTestCase):
    def setUp(self):
        super(DeployTestCase, self).setUp()
        self.site.config.set('aws-bucket-name', 'website')

    def test_simple_deploy(self):
        """
        Test our file deployment

        - Uploaded
        - Have the correct name
        - Compressed (if useful) -- #TODO
        - Publicly readable
        """
        bucket_name = self.site.config.get("aws-bucket-name")

        payload = "\x01" * 1000 + "\x02" * 1000  # Will compress very well
        payload = payload.encode('utf-8')

        with io.FileIO(os.path.join(self.site.static_path, "static.css"), 'w') as f:
            f.write(payload)

        self.site.upload()

        puts = [req for req in self.connection_factory.requests if req.method == "PUT"]

        # How many files did we upload?
        self.assertEqual(1, len(puts))
        put = puts[0]

        # What file did we upload?
        self.assertEqual("/{0}/static/static.css".format(bucket_name), put.url)

        # Where the AWS standard headers correct?
        self.assertEqual("public-read", put.headers["x-amz-acl"])
        self.assertEqual("gzip", put.headers["content-encoding"])
        self.assertEqual("max-age={0}".format(BaseFile.DEFAULT_CACHE_EXPIRATION), put.headers["cache-control"])
        # We just have to check that the max age is set. Another test (test_deploy) checks that this value can be
        # changed using plugins

        # Did we use the correct access key?
        self.assertEqual("AWS 123", put.headers["authorization"].split(':')[0])

        # Did we talk to the right host?
        self.assertEqual("s3.amazonaws.com", put.connection.host)

        # Are the file contents correct?
        compressed = gzip.GzipFile(fileobj=BytesIO(put.body), mode="r")
        self.assertEqual(payload, compressed.read())

    def test_compression(self):
        compress_extensions = ["yes", "html"]
        payload = "\x01" * 1000 + "\x02" * 1000  # Will compress very well
        payload = payload.encode('utf-8')

        self.site.compress_extensions = compress_extensions

        with io.FileIO(os.path.join(self.site.static_path, "static.yes"), 'wb') as f:
            f.write(payload)

        with io.FileIO(os.path.join(self.site.static_path, "static.no"), 'wb') as f:
            f.write(payload)

        self.site.upload()

        puts = [req for req in self.connection_factory.requests if req.method == "PUT"]

        self.assertEqual(2, len(puts))
        compressed = 0
        for put in puts:
            if put.url.rsplit(".", 1)[1] in compress_extensions:
                self.assertEqual("gzip", put.headers["content-encoding"])
                compressed += 1
            else:
                self.assertIsNone(put.headers.get("content-encoding"))

        self.assertEqual(1, compressed)



#coding:utf-8
import logging
import socket

from boto.exception import S3ResponseError

from cactus.deployment.file import BaseFile
from cactus.utils.helpers import CaseInsensitiveDict
from cactus.utils.network import retry
from cactus.utils.url import getURLHeaders
from cactus.utils import ipc


logger = logging.getLogger(__name__)


class S3File(BaseFile):
    def __init__(self, engine, path):
        super(S3File, self).__init__(engine, path)
        self.extra_headers = CaseInsensitiveDict()

    def get_headers(self):
        headers = CaseInsensitiveDict()
        headers['Cache-Control'] = 'max-age={0}'.format(self.cache_control)
        if self.content_encoding is not None:
            headers['Content-Encoding'] = self.content_encoding
        headers.update(self.extra_headers)
        return headers

    def remote_url(self):
        return 'http://%s/%s' % (self.engine.site.config.get('aws-bucket-website'), self.url)

    @retry((S3ResponseError, socket.error, socket.timeout), tries=5, delay=1, backoff=2)
    def remote_changed(self):
        remote_headers = dict((k, v.strip('"')) for k, v in getURLHeaders(self.remote_url()).items())
        local_headers = self.get_headers()
        local_headers['etag'] = self.payload_checksum
        for k, v in local_headers.items():  # Don't check AWS' own headers.
            if remote_headers.get(k) != v:
                return True
        return False

    @retry((S3ResponseError, socket.error, socket.timeout), tries=5, delay=1, backoff=2)
    def do_upload(self):

        kbConstant = (1024 * 100)

        progressCallbackCount = len(self.payload()) / kbConstant

        def progressCallback(current, total):

            if current == 0 or total == 0:
                return

            self.total_bytes_uploaded = current

            logger.info('+ %s upload progress %.1f%%',
                self.url, float(current) / float(total) * 100)

            # logger.warning("deploy.progress %s", self.engine.progress())

            ipc.signal("deploy.progress", {
                "progress": self.engine.progress(),
                "fileName": self.path
            })

        if len(self.payload()) < kbConstant:
            progressCallback = None
            progressCallbackCount = None

        key = self.engine.bucket.new_key(self.url)

        if self.content_type:
            key.content_type = self.content_type  # We don't it need before (local headers only)
        key.md5 = self.payload_checksum   # In case of a flaky network
        key.set_contents_from_string(self.payload(),
            headers=self.get_headers(),
            policy='public-read',
            cb=progressCallback,
            num_cb=progressCallbackCount)



#coding:utf-8

class TestPluginMethod(object):
    def __init__(self, fn=None):
        self.calls = []
        self.fn = fn

    def __call__(self, *args, **kwargs):
        self.calls.append({'args': args, 'kwargs': kwargs})
        if self.fn is not None:
            return self.fn(*args, **kwargs)


preBuildPage = TestPluginMethod(lambda page, context, data: (context, data,))  # site, page, context, data
postBuildPage = TestPluginMethod()  # page / site, page, context, data
preBuild = TestPluginMethod()  # site
postBuild = TestPluginMethod()  # site
preDeploy = TestPluginMethod()  # site
postDeploy = TestPluginMethod()  # site
preDeployFile = TestPluginMethod()  # file

ORDER = 2



#coding:utf-8
import os

from cactus.tests import SiteTestCase


class TestPrettyURLS(SiteTestCase):
    def get_config_for_test(self):
        return {"prettify": True}

    def setUp(self):
        super(TestPrettyURLS, self).setUp()

        open(os.path.join(self.path, 'pages', 'test.html'), 'w')
        subfolder = os.path.join(self.path, 'pages', 'folder')
        os.makedirs(subfolder)
        open(os.path.join(subfolder, 'index.html'), 'w')
        open(os.path.join(subfolder, 'page.html'), 'w')


        self.site.build()

    def test_get_path(self):
        """
        Test that URL rewriting makes pretty links.
        """
        self.assertEqual(self.site.get_url_for_page('/index.html'), '/')
        self.assertEqual(self.site.get_url_for_page('/test.html'), '/test/')
        self.assertEqual(self.site.get_url_for_page('/folder/index.html'), '/folder/')
        self.assertEqual(self.site.get_url_for_page('/folder/page.html'), '/folder/page/')

    def test_build_page(self):
        """
        Check that we rewrite paths for .html files.
        """
        self.assertFileExists(os.path.join(self.site.build_path, 'index.html'))
        self.assertFileExists(os.path.join(self.site.build_path, 'test', 'index.html'))
        self.assertFileExists(os.path.join(self.site.build_path, 'folder', 'index.html'))
        self.assertFileExists(os.path.join(self.site.build_path, 'folder', 'page', 'index.html'))
        self.assertRaises(IOError, open, os.path.join(self.path, '.build', 'test.html'))

    def test_ignore_non_html(self):
        """
        Check that we don't rewrite paths for .txt files.
        """
        self.assertFileExists(os.path.join(self.site.build_path, 'sitemap.xml'))
        self.assertFileExists(os.path.join(self.site.build_path, 'robots.txt'))



import os
import logging


def signal(signal, data=None):
    if data is None:
        data = {}

    if not os.environ.get('DESKTOPAPP'):
        return

    data["signal"] = signal
    logging.warning("", data)



#coding:utf-8
import os
import shutil
import tarfile
import tempfile
import unittest
import zipfile
import threading
import random

from six.moves import BaseHTTPServer, SimpleHTTPServer, xrange

from cactus.bootstrap import bootstrap
from cactus.tests import BaseBootstrappedTestCase
from cactus.utils.filesystem import fileList


def ArchiveServerHandlerFactory(archive_path):
    class ArchiveHandler(SimpleHTTPServer.SimpleHTTPRequestHandler):
        def do_GET(self):
            """
            Reply with the archive.
            """
            self.send_response(200)
            self.end_headers()
            with open(archive_path, 'rb') as f:
                self.copyfile(f, self.wfile)

        def log_request(self, code='-', size='-'):
            """
            Discard log requests to clear up test output.
            """
            return

    return ArchiveHandler


class TestFolderBootstrap(BaseBootstrappedTestCase):
    def test_bootstrap(self):
        self.assertEqual(
            sorted(fileList(self.path, relative=True)),
            sorted(fileList("cactus/tests/data/skeleton", relative=True)),
        )


class TestCactusPackageBootstrap(BaseBootstrappedTestCase):
    def setUp(self):
        self.test_dir = tempfile.mkdtemp()
        self.path = os.path.join(self.test_dir, 'test')
        self.clear_django_settings()
        bootstrap(self.path)

    def test_bootstrap(self):
        self.assertEqual(
            sorted(fileList(self.path, relative=True)),
            sorted(fileList("cactus/skeleton", relative=True)),
        )


class BaseTestArchiveBootstrap(object):
    def setUp(self):
        self.test_dir = tempfile.mkdtemp()
        self.path = os.path.join(self.test_dir, 'test')
        self.skeleton_path = "cactus/skeleton"
        self.archive_path = os.path.join(self.test_dir, "archive")

        with open(self.archive_path, "wb") as f:
            self.make_archive(f)

    def make_archive(self, f):
        raise NotImplementedError()

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def test_file(self):
        # Test
        bootstrap(self.path, self.archive_path)

        self.assertEqual(
            sorted(fileList(self.path, relative=True)),
            sorted(fileList(self.skeleton_path, relative=True)),
        )

    def test_url(self):
        archive_path = self.archive_path

        port = random.choice(xrange(7000, 10000))

        server_address = ("127.0.0.1", port)
        httpd = BaseHTTPServer.HTTPServer(server_address, ArchiveServerHandlerFactory(archive_path))
        t = threading.Thread(target=httpd.serve_forever)
        t.start()

        bootstrap(self.path, "http://127.0.0.1:%s" % port)
        httpd.shutdown()

        self.assertEqual(
            sorted(fileList(self.path, relative=True)),
            sorted(fileList(self.skeleton_path, relative=True)),
        )


class ZIPTestArchiveBootstrap(BaseTestArchiveBootstrap, unittest.TestCase):
    """
    Test ZIP archive support
    """
    def make_archive(self, f):
        archive = zipfile.ZipFile(f, mode="w")
        for resource in fileList(self.skeleton_path, relative=True):
            archive.write(os.path.join(self.skeleton_path, resource), resource)
        archive.close()


class TARTestArchiveBootstrap(BaseTestArchiveBootstrap, unittest.TestCase):
    """
    Test TAR archive support
    """
    def make_archive(self, f):
        archive = tarfile.open(f.name, fileobj=f, mode="w")
        for resource in fileList(self.skeleton_path, relative=True):
            archive.add(os.path.join(self.skeleton_path, resource), resource)
        archive.close()



#coding:utf-8
import os

from cactus.tests import SiteTestCase


class TestLegacyContext(SiteTestCase):
    def setUp(self):
        super(TestLegacyContext, self).setUp()
        os.mkdir(os.path.join(self.site.page_path, "test"))

        with open(os.path.join(self.site.page_path, "static.html"), "w") as f:
            f.write("{{ STATIC_URL }}")

        with open(os.path.join(self.site.page_path, "test", "static.html"), "w") as f:
            f.write("{{ STATIC_URL }}")

        with open(os.path.join(self.site.page_path, "root.html"), "w") as f:
            f.write("{{ ROOT_URL }}")

        with open(os.path.join(self.site.page_path, "test", "root.html"), "w") as f:
            f.write("{{ ROOT_URL }}")

        with open(os.path.join(self.site.page_path, "page.html"), "w") as f:
            f.write("{{ PAGE_URL }}")

    def test_context(self):
        self.site.build()

        with open(os.path.join(self.site.build_path, "static.html")) as f:
            self.assertEqual(f.read(), "./static")

        with open(os.path.join(self.site.build_path, "test", "static.html")) as f:
            self.assertEqual(f.read(), "../static")

        with open(os.path.join(self.site.build_path, "root.html")) as f:
            self.assertEqual(f.read(), ".")

        with open(os.path.join(self.site.build_path, "test", "root.html")) as f:
            self.assertEqual(f.read(), "..")

        with open(os.path.join(self.site.build_path, "page.html")) as f:
            self.assertEqual(f.read(), "page.html")

    def test_pretty_urls(self):
        self.site.prettify_urls = True

        self.site.build()

        with open(os.path.join(self.site.build_path, "test", "static", "index.html")) as f:
            self.assertEqual(f.read(), "../../static")

        with open(os.path.join(self.site.build_path, "root", "index.html")) as f:
            self.assertEqual(f.read(), "..")

        with open(os.path.join(self.site.build_path, "test", "root", "index.html")) as f:
            self.assertEqual(f.read(), "../..")

        with open(os.path.join(self.site.build_path, "page", "index.html")) as f:
            self.assertEqual(f.read(), "page/")



#coding:utf-8
import webbrowser

from oauth2client.client import OAuth2WebServerFlow, FlowExchangeError
from oauth2client.contrib.keyring_storage import Storage

from cactus.deployment.gcs import CACTUS_CLIENT_ID, CACTUS_CLIENT_SECRET, CACTUS_REQUIRED_SCOPE, LOCAL_REDIRECT_URI
from cactus.exceptions import InvalidCredentials


class GCSCredentialsManager(object):
    def __init__(self, engine):
        self.engine = engine  #TODO: Only pass those things that are needed?
        self.credentials = None

    def get_storage(self):
        return Storage("cactus/gcs", self.engine.bucket_name)  #TODO: Not a great key, but do we want to ask for email?

    def get_credentials(self):

        if self.credentials is not None:
            return self.credentials

        self.credentials = self.get_storage().get()

        if self.credentials is None:
            flow = OAuth2WebServerFlow(
                client_id=CACTUS_CLIENT_ID,
                client_secret=CACTUS_CLIENT_SECRET,
                scope=CACTUS_REQUIRED_SCOPE,
                redirect_uri=LOCAL_REDIRECT_URI
            )

            auth_uri = flow.step1_get_authorize_url()
            webbrowser.open(auth_uri)  #TODO: Actually print the URL...
            code = self.engine.site.ui.prompt('Please enter the authorization code')

            try:
                self.credentials = flow.step2_exchange(code)  #TODO: Catch invalid grant
            except FlowExchangeError:
                raise InvalidCredentials("The authorization did not match.")

        return self.credentials

    def save_credentials(self):
        assert self.credentials is not None, "You did not set credentials before saving them"  #TODO: That's still bad
        self.get_storage().put(self.credentials)



#coding:utf-8
import os
import shutil

from cactus.plugin.loader import CustomPluginsLoader
from cactus.plugin.manager import PluginManager
from cactus.tests import SiteTestCase


class TestPluginLoader(SiteTestCase):
    def setUp(self):
        super(TestPluginLoader, self).setUp()
        self.site.plugin_manager = PluginManager(self.site, [CustomPluginsLoader(self.site.path)])
        shutil.rmtree(self.site.plugin_path)
        os.makedirs(self.site.plugin_path)

    def _load_test_plugin(self, plugin, to_filename):
        src_path = os.path.join('cactus', 'tests', 'data', 'plugins', plugin)
        dst_path = os.path.join(self.site.plugin_path, to_filename)
        shutil.copy(src_path, dst_path)
        self.site.plugin_manager.reload()

    def test_ignore_disabled(self):
        self._load_test_plugin('test.py', 'test.disabled.py')
        self.assertEqual([], [p for p in self.site.plugin_manager.plugins if not p.builtin])

    def test_load_plugin(self):
        self._load_test_plugin('test.py', 'test.py')

        plugins = self.site.plugin_manager.plugins
        self.assertEqual(1, len(plugins ))
        self.assertEqual('plugin_test', plugins[0].plugin_name)
        self.assertEqual(2, plugins[0].ORDER)

    def test_defaults(self):
        """
        Check that defaults get initialized
        """
        self._load_test_plugin('empty.py', 'empty.py')
        plugins = self.site.plugin_manager.plugins

        plugin = plugins[0]
        self.assert_(hasattr(plugin, 'preBuild'))
        self.assert_(hasattr(plugin, 'postBuild'))
        self.assertEqual(-1, plugin.ORDER)

    def test_call(self):
        """
        Check that plugins get called
        """
        self._load_test_plugin('test.py', 'call.py')
        plugins = self.site.plugin_manager.plugins

        plugin = plugins[0]

        self.assertEqual('plugin_call', plugin.plugin_name)  # Just to check we're looking at the right one.

        self.site.build()

        # preBuild
        self.assertEqual(1, len(plugin.preBuild.calls))
        self.assertEqual((self.site,), plugin.preBuild.calls[0]['args'])

        # preBuildPage
        self.assertEqual(len(self.site.pages()), len(plugin.preBuildPage.calls))
        for call in plugin.preBuildPage.calls:
            self.assertIn(len(call['args']), (3, 4))

        # postBuildPage
        self.assertEqual(len(self.site.pages()), len(plugin.postBuildPage.calls))

        #postBuild
        self.assertEqual(1, len(plugin.postBuild.calls))
        self.assertEqual((self.site,), plugin.postBuild.calls[0]['args'])



#coding:utf-8
import io
import base64
import socket

from apiclient.http import MediaIoBaseUpload
from apiclient.errors import HttpError
from cactus.deployment.file import BaseFile
from cactus.utils.network import retry


class GCSFile(BaseFile):
    def get_metadata(self):
        """
        Generate metadata for the upload.
        Note: we don't set the etag, since the GCS API does not accept what we set
        """
        metadata = {
            "acl": [{"entity": "allUsers", "role": "READER"},],
            "md5Hash": base64.b64encode(self.payload_checksum.decode('hex')),
            "contentType": self.content_type,  # Given twice...
            "cacheControl": unicode(self.cache_control)  # That's what GCS will return
        }

        if self.content_encoding is not None:
            metadata['contentEncoding'] = self.content_encoding

        return metadata

    def remote_changed(self):
        """
        Compare each piece of metadata that we're setting with the one that's stored remotely
        If one's different, upload again.

        :rtype: bool
        """
        resource = self.engine.get_connection().objects()
        req = resource.get(bucket=self.engine.bucket_name, object=self.url)

        try:
            remote_metadata = req.execute()
        except HttpError as e:
            if e.resp.status == 404:
                return True
            raise

        ignore_metadata = ["acl"] # We can't control what we'll retrieve TODO: do the best we can do!

        for k, v in self.get_metadata().items():
            if k not in ignore_metadata and remote_metadata.get(k) != v:
                return True
        return False

    @retry((socket.error,), tries=5, delay=3, backoff=2)
    def do_upload(self):
        resource = self.engine.get_connection().objects()

        stream = io.BytesIO(self.payload())
        upload = MediaIoBaseUpload(stream, mimetype=self.content_type)

        req = resource.insert(
            bucket=self.engine.bucket_name,
            name=self.url,
            body=self.get_metadata(),
            media_body=upload,
        )

        req.execute()



#coding:utf-8
import shutil
import tempfile
import unittest

from cactus.plugin.manager import PluginManager
from cactus.utils.parallel import PARALLEL_DISABLED
from cactus.config.fallback import ConfigFallback
from cactus.deployment.engine import BaseDeploymentEngine
from cactus.deployment.file import BaseFile


class DummyUI(object):
    """
    A class to mock the site's UI and have it return what we need
    """
    def __init__(self, bucket_name="test-bucket", create_bucket=True):
        self.bucket_name = bucket_name
        self.create_bucket = create_bucket

        self.asked_name = 0
        self.asked_create = 0

    def prompt_normalized(self, q):
        self.asked_name +=1
        return self.bucket_name

    def prompt_yes_no(self, q):
        self.asked_create += 1
        return self.create_bucket


class DummyConfig(ConfigFallback):
    """
    A ConfigFallback that remembers whether we saved
    """
    def write(self):
        super(DummyConfig, self).write()
        self.saved = True


class DummyPluginManager(PluginManager):
    """
    Doesn't do anything
    """
    def call(self, method, *args, **kwargs):
        """
        Trap the call
        """
        pass


class DummySite(object):
    """
    Something that pretends to be a site, as far as the deployment engine knows
    """
    _parallel = PARALLEL_DISABLED

    def __init__(self, path, ui):
        self.build_path = path
        self.config = ConfigFallback()
        self.ui = ui
        self.plugin_manager = DummyPluginManager(self, [])
        self.compress_extensions = []


class DummyCredentialsManager(object):
    """
    Something that looks like a CredentialsManager, as far as we know.
    """
    _username = "123"
    _password = "456"

    def __init__(self, engine):
        self.engine = engine
        self.saved = False

    def get_credentials(self):
        return self._username, self ._password

    def save_credentials(self):
        self.saved = True


class DummyFile(BaseFile):
    """
    A fake file class we can extend to test things
    """
    def __init__(self, engine, path):
        super(DummyFile, self).__init__(engine, path)
        self.engine.created_files.append(self)
        self.remote_changed_calls = 0
        self.do_upload_calls = 0

    def remote_changed(self):
        self.remote_changed_calls += 1
        return True

    def do_upload(self):
        self.do_upload_calls += 1
        pass


class DummyDeploymentEngine(BaseDeploymentEngine):
    """
    A deployment engine we can extend
    """
    FileClass = DummyFile
    CredentialsManagerClass = DummyCredentialsManager

    config_bucket_name = "test-conf-entry"
    config_bucket_website = "test-conf-entry-website"

    def __init__(self, site):
        super(DummyDeploymentEngine, self).__init__(site)
        self.get_bucket_calls = 0
        self.create_bucket_calls = 0
        self.get_website_calls = 0
        self.created_files = []

    def _create_connection(self):
        pass

    def get_bucket(self):
        self.get_bucket_calls += 1
        if self.create_bucket_calls:
            return "test-bucket-obj"
        return None

    def create_bucket(self):
        self.create_bucket_calls += 1
        return "test-bucket-obj"

    def get_website_endpoint(self):
        self.get_website_calls += 1
        return "http://test-bucket.com"


class BaseDeploymentTestCase(unittest.TestCase):
    def setUp(self):
        self.test_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.test_dir)



#coding:utf-8
import logging

from cactus.config.fallback import ConfigFallback
from cactus.config.file import ConfigFile


logger = logging.getLogger(__name__)


class ConfigRouter(object):
    """
    A router to manage a series of config files.
    """

    def __init__(self, paths):
        """
        Load all the config files passed.
        Make sure not to load the same one twice
        """
        self.configs = []

        loaded_paths = set()
        for path in paths:
            if path not in loaded_paths:
                self.configs.append(ConfigFile(path))
                loaded_paths.add(path)

        self.configs.append(ConfigFallback())

        logger.debug("Loaded configs: %s", ', '.join([config.path for config in self.configs]))


    def _get_nested(self, key, default=None):
        assert isinstance(default, dict)  # Don't shoot yourself in the foot.

        output = {}
        for config in reversed(self.configs):
            output.update(config.get(key, default))
            logger.debug("Retrieving %s from %s", key, config.path)

        return output

    def _get_first(self, key, default=None):
        for config in self.configs:
            if config.has_key(key):
                logger.debug("Retrieved %s from %s", key, config.path)
                return config.get(key)

        return default

    def get(self, key, default=None, nested=False):  #TODO: Mutable default.. copy?
        """
        Retrieve a config key from the first config that has it.
        Return default if no config has it.
        """
        logger.debug("Searching for %s (nested:%s)", key, nested)
        if nested:
            return self._get_nested(key, default)
        else:
            return self._get_first(key, default)


    def set(self, key, value):
        """
        Write a config key from the first config that has it.

        If none do, write it to the first one.
        """
        assert self.configs  # There should always be at least a fallback config

        write_to = None

        for config in self.configs:
            if config.has_key(key):
                write_to = config
        if write_to is None:
            write_to = self.configs[0]

        write_to.set(key, value)
        logger.debug("Set %s in %s", key, write_to.path)

    def write(self):
        """
        Write the config files to the filesystem.
        """
        for config in self.configs:
            config.write()



#coding:utf-8
import os
import re

from cactus.deployment.s3.engine import S3DeploymentEngine
from cactus.utils.helpers import checksum

from cactus.tests.integration import IntegrationTestCase, DebugHTTPSConnectionFactory, BaseTestHTTPConnection, \
    TestHTTPResponse


class DummyAWSCredentialsManager(object):
    def __init__(self, site):
        self.site = site

    def get_credentials(self):
        return "123", "abc"

    def save_credentials(self):
        pass


class S3TestHTTPConnection(BaseTestHTTPConnection):

    def handle_request(self, request):
        if request.method == "GET":
            # List buckets
            if request.path == "/":
                if request.params == {}:
                    return self.list_buckets()
            # Request for just one bucket (like /bucket/ - regex is not perfect but should do for here)
            if re.match(r"/[a-z1-9\-.]+/", request.path):
                if "location" in request.params:
                    return self.location()

        if request.method == "PUT":
            if request.path == "/":
                return TestHTTPResponse(200)
            return self.put_object(request)

        raise Exception("Unsupported request {0} {1}".format(request.method, request.url))

    def _serve_data(self, name):
        with open(os.path.join("cactus/tests/integration/s3/data", name)) as f:
            return TestHTTPResponse(200, body=f.read())

    def list_buckets(self):
        return self._serve_data("buckets.xml")

    def location(self):
        return self._serve_data("location.xml")

    def put_object(self, req):
        return TestHTTPResponse(200, headers={"ETag":'"{0}"'.format(checksum(req.body))})


class S3IntegrationTestCase(IntegrationTestCase):
    def get_deployment_engine_class(self):
        # Create a connection factory
        self.connection_factory = DebugHTTPSConnectionFactory(S3TestHTTPConnection)

        class TestS3DeploymentEngine(S3DeploymentEngine):
            _s3_https_connection_factory = (self.connection_factory, ())
            CredentialsManagerClass = DummyAWSCredentialsManager

        return TestS3DeploymentEngine

    def get_credentials_manager_class(self):
        return DummyAWSCredentialsManager



import os
import mimetypes

MIMETYPE_MAP = {
    '.js':   'text/javascript',
    '.mov':  'video/quicktime',
    '.mp4':  'video/mp4',
    '.m4v':  'video/x-m4v',
    '.3gp':  'video/3gpp',
    '.woff': 'application/font-woff',
    '.eot':  'application/vnd.ms-fontobject',
    '.ttf':  'application/x-font-truetype',
    '.otf':  'application/x-font-opentype',
    '.svg':  'image/svg+xml',
}

MIMETYPE_DEFAULT = 'application/octet-stream'

def guess(path):

    if not path:
        return MIMETYPE_DEFAULT

    base, ext = os.path.splitext(path)

    if ext.lower() in MIMETYPE_MAP:
        return MIMETYPE_MAP[ext.lower()]

    mime_type, encoding = mimetypes.guess_type(path)

    if mime_type:
        return mime_type

    return MIMETYPE_DEFAULT



import posixpath
import os


def pkg_walk(package, top):
    # we get names tuple names where names[0] is current folder, names[1] is list of dirs and names[2] is list of files
    cur_dir, dirs, non_dirs = next(os.walk(os.path.join(package, top)))
    yield top, dirs, non_dirs
    # loop through dirs to do the same thing as we did before
    for name in dirs:
        new_path = posixpath.join(top, name)
        for out in pkg_walk(package, new_path):
            yield out



#coding:utf-8
from cactus.tests.integration.s3 import S3IntegrationTestCase


class BucketTestCase(S3IntegrationTestCase):
    def test_create_bucket(self):
        """
        Test that we properly create a bucket in AWS
        """
        self.site.deployment_engine.bucket_name = "new"
        bucket = self.site.deployment_engine.create_bucket()

        self.assertEqual("new", bucket.name)

        self.assertEqual(2, len(self.connection_factory.requests))
        new, configure = self.connection_factory.requests

        self.assertEqual("/new/", new.path)
        self.assertEqual("PUT", new.method)

        self.assertEqual("/new/?website", configure.url)
        self.assertEqual("PUT", configure.method)

        for req in (new, configure):
            self.assertEqual("s3.amazonaws.com", req.connection.host)

    def test_list_buckets(self):
        """
        Test that we retrieve the correct list of buckets from AWS
        """
        buckets = self.site.deployment_engine._get_buckets()

        # Check that the right buckets are returned
        bucket_names = [bucket.name for bucket in buckets]
        self.assertEqual(sorted(bucket_names), sorted(["website", "other"]))

        # Check that we only made one connection to list buckets
        self.assertEqual(1, len(self.connection_factory.requests))
        req = self.connection_factory.requests[0]
        self.assertEqual("GET", req.method)
        self.assertEqual("/", req.path)

    def test_get_bucket(self):
        """
        Test that we access the correct bucket in AWS
        """
        self.site.deployment_engine.bucket_name = "other"
        bucket = self.site.deployment_engine.get_bucket()

        # Check that just one bucket was retried
        self.assertEqual("other", bucket.name)

        # Check that we only made one connection to list buckets
        self.assertEqual(1, len(self.connection_factory.requests))
        req = self.connection_factory.requests[0]
        self.assertEqual("GET", req.method)
        self.assertEqual("/", req.path)



#coding:utf-8
import six
import inspect

# Adapted from: http://kbyanc.blogspot.com/2007/07/python-more-generic-getargspec.html


FUNC_OBJ_ATTR = "__func__" if six.PY3 else "im_func"


def getargspec(obj):
    """
    Get the names and default values of a callable's
       arguments

    A tuple of four things is returned: (args, varargs,
    varkw, defaults).
      - args is a list of the argument names (it may
        contain nested lists).
      - varargs and varkw are the names of the * and
        ** arguments or None.
      - defaults is a tuple of default argument values
        or None if there are no default arguments; if
        this tuple has n elements, they correspond to
        the last n elements listed in args.

    Unlike inspect.getargspec(), can return argument
    specification for functions, methods, callable
    objects, and classes.  Does not support builtin
    functions or methods.
    """
    if not callable(obj):
        raise TypeError("%s is not callable" % type(obj))
    try:
        if inspect.isfunction(obj):
            return inspect.getfullargspec(obj)
        elif hasattr(obj, FUNC_OBJ_ATTR):
            # For methods or classmethods drop the first
            # argument from the returned list because
            # python supplies that automatically for us.
            # Note that this differs from what
            # inspect.getargspec() returns for methods.
            # NB: We use im_func so we work with
            #     instancemethod objects also.
            spec = inspect.getargspec(getattr(obj, FUNC_OBJ_ATTR))
            return inspect.ArgSpec(spec.args[:1], spec.varargs, spec.keywords, spec.defaults)
        elif inspect.isclass(obj):
            return getargspec(obj.__init__)
        elif isinstance(obj, object):
            # We already know the instance is callable,
            # so it must have a __call__ method defined.
            # Return the arguments it expects.
            return getargspec(obj.__call__)
    except NotImplementedError:
        # If a nested call to our own getargspec()
        # raises NotImplementedError, re-raise the
        # exception with the real object type to make
        # the error message more meaningful (the caller
        # only knows what they passed us; they shouldn't
        # care what aspect(s) of that object we actually
        # examined).
        pass
    raise NotImplementedError("do not know how to get argument list for %s" % type(obj))



#coding:utf-8
import subprocess
from cactus.static.external import External


class ClosureJSOptimizer(External):
    supported_extensions = ('js',)
    output_extension = 'js'

    def _run(self):
        subprocess.call([
            'closure-compiler',
            '--js', self.src,
            '--js_output_file', self.dst,
            '--compilation_level', 'SIMPLE_OPTIMIZATIONS'
        ])



import os
import sys
import pipes
import shutil
import subprocess

from cactus.utils.filesystem import fileList

"""
This plugin uses pyScss to translate sass files to css

Install:

sudo easy_install pyScss

"""

try:
    from scss import Scss
except:
    sys.exit("Could not find pyScss, please install: sudo easy_install pyScss")


CSS_PATH = 'static/css'

for path in fileList(CSS_PATH):

    if not path.endswith('.scss'):
        continue

    with open(path, 'r') as f:
        data = f.read()

    css = Scss().compile(data)

    with open(path.replace('.scss', '.css'), 'w') as f:
        f.write(css)



#coding:utf-8
import unittest

from cactus.mime import guess


class MimeTestCase(unittest.TestCase):
    def test_mime_eot(self):
        font_path = "test/font.eot"
        self.assertFalse(guess(font_path) is None)

    def test_mime_dummy(self):
        # Make sure we never return a None mime type!
        path = "test/format.thisisnomime"
        self.assertFalse(guess(path) is None)



import subprocess
import platform

s1 = """
set hostLists to %s
tell application "Google Chrome"
    set windowsList to windows as list
    repeat with currWindow in windowsList
        set tabsList to currWindow's tabs as list
        repeat with currTab in tabsList
            repeat with currentHost in hostLists
                if currentHost is in currTab's URL then execute currTab javascript "%s"
            end repeat
        end repeat
    end repeat
end tell
"""

s2 = """
set hostLists to %s
tell application "Safari"
    if (count of windows) is greater than 0 then
        set windowsList to windows as list
        repeat with currWindow in windowsList
            set tabsList to currWindow's tabs as list
            repeat with currTab in tabsList
                repeat with currentHost in hostLists
                    if currentHost is in currTab's URL then
                        tell currTab to do JavaScript "%s"
                    end if
                end repeat
            end repeat
        end repeat
    end if
end tell
"""

s3 = """
window.location.reload()
"""

s4 = """
(function() {
    function updateQueryStringParameter(uri, key, value) {

        var re = new RegExp('([?|&])' + key + '=.*?(&|$)', 'i');
        separator = uri.indexOf('?') !== -1 ? '&' : '?';

        if (uri.match(re)) {
            return uri.replace(re, '$1' + separator + key + '=' + value + '$2');
        } else {
            return uri + separator + key + '=' + value;
        }
    }

    var links = document.getElementsByTagName('link');

    for (var i = 0; i < links.length;i++) {

        var link = links[i];

        if (link.rel === 'stylesheet') {

            // Don't reload external urls, they likely did not change
            if (
                link.href.indexOf('127.0.0.1') == -1 && 
                link.href.indexOf('localhost') == -1 &&
                link.href.indexOf('0.0.0.0') == -1) {
                continue;
            }

            var updatedLink = updateQueryStringParameter(link.href, 'cactus.reload', new Date().getTime());

            // This is really hacky, but needed because the regex gets magically broken by piping it
            // through applescript. This replaces the first occurence of ? with & if there was no &.
            if (updatedLink.indexOf('?') == -1) {
                updatedLink = updatedLink.replace('&', '?');
            }

            link.href = updatedLink;
        }
    }
})()
"""

def applescript(input):

    # Bail if we're not on mac os for now
    if platform.system() != "Darwin":
        return

    command = "osascript<<END%sEND" % input

    try:
        return subprocess.check_call(command,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=True)
    except Exception as e:
        pass

    # return subprocess.check_output(command, shell=True)

def _insertJavascript(urlMatch, js):

    apps = appsRunning(['Safari', 'Google Chrome'])

    # urlMatch is a list that needs to be converted to applescript format
    # ["a", "b"] -> {"a", "b"}
    urlMatch = "{\"" + "\",\"".join(urlMatch) + "\"}"

    if apps['Google Chrome']:
        applescript(s1 % (urlMatch, js))

    if apps['Safari']:
        applescript(s2 % (urlMatch, js))

def browserReload(url):
    _insertJavascript(url, s3)

def browserReloadCSS(url):
    _insertJavascript(url, s4)

def appsRunning(l):
    psdata = subprocess.check_output(['ps aux'], shell=True)
    retval = {}
    for app in l: retval[app] = app in psdata
    return retval



import sys
import os
import codecs

# How to:
#     * Install hamlpy (https://github.com/jessemiller/HamlPy)
#     * .haml files will compiled to .html files


from hamlpy.hamlpy import Compiler
from cactus.utils.filesystem import fileList

CLEANUP = []

def preBuild(site):
    for path in fileList(site.paths['pages']):

        #only file ends with haml
        if not path.endswith('.haml'):
            continue

        #read the lines
        haml_lines = codecs.open(path, 'r', encoding='utf-8').read().splitlines()

        #compile haml to html
        compiler = Compiler()
        output = compiler.process_lines(haml_lines)

        #replace path
        outPath = path.replace('.haml', '.html')

        #write the html file
        with open(outPath,'w') as f:
            f.write(output)

        CLEANUP.append(outPath)


def postBuild(site):
    global CLEANUP
    for path in CLEANUP:
        print path
        os.remove(path)
    CLEANUP = []



#coding:utf-8

def preBuildPage(page, context, data):
    """
    Updates the context of the page to include: the page itself as {{ CURRENT_PAGE }}
    """

    # This will run for each page that Cactus renders.
    # Any changes you make to context will be passed to the template renderer for this page.

    extra = {
        "CURRENT_PAGE": page
        # Add your own dynamic context elements here!
    }

    context.update(extra)
    return context, data



#coding:utf-8
import logging
import threading

import httplib2
import apiclient.discovery
import apiclient.errors

from cactus.deployment.engine import BaseDeploymentEngine
from cactus.deployment.gcs.auth import GCSCredentialsManager
from cactus.deployment.gcs.file import GCSFile
from cactus.exceptions import InvalidCredentials


class GCSDeploymentEngine(BaseDeploymentEngine):
    FileClass = GCSFile
    CredentialsManagerClass = GCSCredentialsManager

    config_bucket_name = "gcs-bucket-name"
    config_bucket_website = "gcs-bucket-website"

    _HTTPClass = httplib2.Http


    def __init__(self, *args, **kwargs):
        super(GCSDeploymentEngine, self).__init__(*args, **kwargs)
        self._service_pool = {}  # We can't share services (they share SSL connections) across processes.

    def get_connection(self):
        """
        Worker threads may not share the same connection
        """
        thread = threading.current_thread()
        ident = thread.ident

        service = self._service_pool.get(ident)
        if service is None:
            credentials = self.credentials_manager.get_credentials()
            http_client = self._HTTPClass()
            credentials.authorize(http_client)
            service = apiclient.discovery.build('storage', 'v1', http=http_client)
            self._service_pool[ident] = service

        return service

    def get_bucket(self):
        req = self.get_connection().buckets().get(bucket=self.bucket_name)
        try:
            return req.execute()
        except apiclient.errors.HttpError as e:
            if e.resp['status'] == '404':
                return None
            raise

    def create_bucket(self):
        project_id = self.site.ui.prompt_normalized('API project identifier?')

        public_acl = {
            "entity": "allUsers",
            "role": "READER",
        }

        body = {
            "name": self.bucket_name,
            "website": {
                "mainPageSuffix": self._index_page,
                "notFoundPage": self._error_page,
            },
            "defaultObjectAcl": [public_acl], #TODO: Not required actually
        }

        self.get_connection().buckets().insert(project=project_id, body=body).execute()
        return self.get_bucket()

    def get_website_endpoint(self):
        return "Unavailable for GCS"



#coding:utf-8
from __future__ import unicode_literals

import os
import io
import hashlib

from cactus.tests import SiteTestCase


class TestFingerPrintingMixin(object):
    fingerprint_extensions = None

    def get_config_for_test(self):
        return {"fingerprint": self.fingerprint_extensions}


class TestFingerprintingOff(TestFingerPrintingMixin, SiteTestCase):
    fingerprint_extensions = []

    def setUp(self):
        super(TestFingerprintingOff, self).setUp()
        self.site.build()

    def test_fingerprinting_off(self):
        """
        Test that fingerprinting can be disabled.
        """
        static = '/static/css/style.css'
        self.assertEqual(self.site.get_url_for_static(static), static)
        self.assertFileExists(os.path.join(self.site.build_path, self.site.get_url_for_static(static)[1:]))


class TestFingerprintingOn(TestFingerPrintingMixin, SiteTestCase):
    fingerprint_extensions = ["css", "js"]

    def setUp(self):
        super(TestFingerprintingOn, self).setUp()
        self.site.build()

    def test_fingerprinting_on(self):
        """
        Test that fingerprinting provides existing URLs.
        """
        static = '/static/css/style.css'
        self.assertNotEqual(self.site.get_url_for_static(static), static)
        self.assertFileExists(os.path.join(self.site.build_path, self.site.get_url_for_static(static)[1:]))

        static = '/static/js/main.js'
        self.assertNotEqual(self.site.get_url_for_static(static), static)
        self.assertFileExists(os.path.join(self.site.build_path, self.site.get_url_for_static(static)[1:]))


class TestFingerprintingSelective(TestFingerPrintingMixin, SiteTestCase):
    fingerprint_extensions = ["css"]

    def setUp(self):
        super(TestFingerprintingSelective, self).setUp()
        self.site.build()

    def test_fingerprinting_selective(self):
        """
        Test that fingerprinting can be restricted to certain filetypes.
        """
        static = '/static/css/style.css'
        self.assertNotEqual(self.site.get_url_for_static(static), static)
        self.assertFileExists(os.path.join(self.site.build_path, self.site.get_url_for_static(static)[1:]))

        static = '/static/js/main.js'
        self.assertEqual(self.site.get_url_for_static(static), static)
        self.assertFileExists(os.path.join(self.site.build_path, self.site.get_url_for_static(static)[1:]))


class TestFingerprintingValues(TestFingerPrintingMixin, SiteTestCase):
    fingerprint_extensions = ["dat"]

    def test_fingerprinting_hashes(self):
        payload = b"\x01" * 131072  # (2**17 bytes = 2 * 65536)
        expected_checksum = hashlib.md5(payload).hexdigest()

        with io.FileIO(os.path.join(self.path, "static", "data.dat"), "w") as f:
            f.write(payload)

        self.site.build()
        self.assertTrue(expected_checksum in self.site.get_url_for_static("/static/data.dat"))



#coding:utf-8
from __future__ import print_function
import sys
import logging
import multiprocessing.pool


PARALLEL_AGGRESSIVE = 2
PARALLEL_CONSERVATIVE = 1
PARALLEL_DISABLED = 0

def multiMap(f, items, workers = 8):

    # Code in GCS engine +/- depends on this being threads
    pool = multiprocessing.pool.ThreadPool(workers)

    # Simple wrapper to provide decent tracebacks
    def wrapper(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except Exception as e:
            import traceback
            print(traceback.format_exc())
            pool.join()
            sys.exit()

    return pool.map(wrapper, items)



import logging


from cactus.listener.polling import PollingListener


logger = logging.getLogger(__name__)


try:
    from cactus.listener.mac import FSEventsListener as Listener
except (ImportError, OSError):
    logger.debug("Failed to load FSEventsListener, falling back to PollingListener", exc_info=True)
    Listener = PollingListener



#coding:utf-8
import json
import logging


logger = logging.getLogger(__name__)


class ConfigFile(object):
    """
    A single config file, as present on the filesystem.
    """
    _dirty = None

    def __init__(self, path):
        self.path = path
        self.load()

    def get(self, key, default=None):
        return self._data.get(key, default)

    def set(self, key, value):
        self._data[key] = value
        self._dirty = True

    def has_key(self, key):
        return key in self._data

    def load(self):
        self._data = {}

        try:
            self._data = json.load(open(self.path))
            self._dirty = False
        except IOError:
            logger.warning("Unable to load configuration at '{0}'. No file found.".format(self.path))
        except ValueError as e:
            logger.error("Unable to load configuration at '{0}'. Invalid JSON caused by: {1}".format(self.path, e))
        except Exception as e:
            logger.exception("Unable to load configuration at '{0}'.".format(self.path))

    def write(self):
        if self._dirty:
            json.dump(self._data, open(self.path, "w"), sort_keys=True, indent=4, separators=(",", ": "))
            self._dirty = False
        logger.debug("Saved configuration at {0}".format(self.path))



import os
import sys
import logging
import threading
import mimetypes
import itertools

import tornado.httpserver
import tornado.websocket
import tornado.ioloop
import tornado.web

logger = logging.getLogger(__name__)

from cactus import mime

TEMPLATES = {}


class WebSocketHandler(tornado.websocket.WebSocketHandler):

    def open(self):
        if self not in self.application._socketHandlers:
            self.application._socketHandlers.append(self)

    def on_close(self):
        if self in self.application._socketHandlers:
            self.application._socketHandlers.remove(self)

    def on_message(self, msg):
        pass


class ShutdownHandler(tornado.web.RequestHandler):
    def post(self, *args, **kwargs):
        ioloop = tornado.ioloop.IOLoop.instance()
        ioloop.add_callback(lambda x: x.stop(), ioloop)


class StaticHandler(tornado.web.StaticFileHandler):

    @classmethod
    def get_append(cls, abspath):

        mime_type, encoding = mimetypes.guess_type(abspath)

        if mime_type == "text/html":
            return TEMPLATES["script"]

        return ""

    @classmethod
    def get_content(cls, abspath, start=None, end=None):
        data = super(StaticHandler, cls).get_content(abspath, start=start, end=end)
        return itertools.chain(data, [cls.get_append(abspath)])

    def get_content_size(self):
        return super(StaticHandler, self).get_content_size() + len(self.get_append(self.absolute_path))

    # Always be not caching
    def should_return_304(self):
        return False

    def get_content_type(self):
        return mime.guess(self.absolute_path)

    def write_error(self, status_code, **kwargs):
        # Special case handling for 404: try to find one of the error pages,
        # which might be in error.html or error/index.html depending on
        # whether prettifying was enabled.
        if status_code == 404:
            for path in ["error.html", "error/index.html"]:
                try:
                    return self.render(path)
                except IOError:
                    continue

        return super(StaticHandler, self).write_error(status_code, **kwargs)

    def log_request(self, handler):
        pass


class StaticSingleFileHandler(tornado.web.RequestHandler):

    def get(self):
        self.set_header("Content-Type", mime.guess("file.js"))
        self.finish(TEMPLATES["js"])


class WebServer(object):

    def __init__(self, path, port=8080):

        self.path = path
        self.port = port

        self.application = tornado.web.Application([
            (r'/_cactus/shutdown', ShutdownHandler),
            (r'/_cactus/ws', WebSocketHandler),
            (r'/_cactus/cactus.js', StaticSingleFileHandler),
            (r'/(.*)', StaticHandler, {'path': self.path, "default_filename": "index.html"}),
        ], template_path=self.path)

        self.application.log_request = lambda x: self._log_request(x)

    def _log_request(self, handler):

        if not isinstance(handler, StaticHandler):
            return

        if handler.get_status() < 400:
            log_method = logging.info
        elif handler.get_status() < 500:
            log_method = logging.warning
        else:
            log_method = logging.error

        log_method("%d %s %s", handler.get_status(), handler.request.method, handler.request.uri)


    def start(self):
        self.application._socketHandlers = []

        self._server = tornado.httpserver.HTTPServer(self.application)
        self._server.listen(self.port)

        tornado.ioloop.IOLoop.instance().start()

    def stop(self):
        pass

    def publish(self, message):
        for ws in self.application._socketHandlers:
            ws.write_message(message)

    def reloadPage(self):
        self.publish("reloadPage")

    def reloadCSS(self):
        self.publish("reloadCSS")


TEMPLATES["script"] = """

<!-- Automatically inserted by Cactus. Needed for auto refresh. This will be gone when you deploy -->
<script src="/_cactus/cactus.js"></script>
"""

TEMPLATES["js"] = """
(function() {

function reloadPage() {
    window.location.reload()
}

function reloadCSS() {
    function updateQueryStringParameter(uri, key, value) {

        var re = new RegExp("([?|&])" + key + "=.*?(&|$)", "i");
        separator = uri.indexOf("?") !== -1 ? "&" : "?";

        if (uri.match(re)) {
            return uri.replace(re, "$1" + separator + key + "=" + value + "$2");
        } else {
            return uri + separator + key + "=" + value;
        };
    };

    var links = document.getElementsByTagName("link");

    for (var i = 0; i < links.length;i++) {

        var link = links[i];

        if (link.rel === "stylesheet") {

            // Don"t reload external urls, they likely did not change
            if (
                link.href.indexOf("127.0.0.1") == -1 &&
                link.href.indexOf("localhost") == -1 &&
                link.href.indexOf("0.0.0.0") == -1 &&
                link.href.indexOf(window.location.host) == -1
                ) {
                continue;
            }

            var updatedLink = updateQueryStringParameter(link.href, "cactus.reload", new Date().getTime());

            if (updatedLink.indexOf("?") == -1) {
                updatedLink = updatedLink.replace("&", "?");
            };

            link.href = updatedLink;
        };
    };
};

function startSocket() {

    var MessageActions = {
        reloadPage: reloadPage,
        reloadCSS: reloadCSS
    };

    var socketUrl = "ws://" + window.location.host + "/_cactus/ws"
    var socket = new WebSocket(socketUrl);

    socket.onmessage = function(e) {
        var key = e.data;

        if (MessageActions.hasOwnProperty(key)) {
            MessageActions[key]()
        };
    };
};

startSocket();

})()
"""



#coding:utf-8
from cactus.contrib.external.closure import ClosureJSOptimizer
from cactus.contrib.external.yui import YUICSSOptimizer


def preBuild(site):
    """
    Registers optimizers as requested by the configuration.
    Be sure to read the plugin to understand and use it.
    """

    # Inspect the site configuration, and retrieve an `optimize` list.
    # This lets you configure optimization selectively.
    # You may want to use one configuration for staging with no optimizations, and one
    # configuration for production, with all optimizations.
    optimize = site.config.get("optimize", [])

    if "js" in optimize:
        # If `js` was found in the `optimize` key, then register our JS optimizer.
        # This uses closure, but you could use cactus.contrib.external.yui.YUIJSOptimizer!
        site.external_manager.register_optimizer(ClosureJSOptimizer)

    if "css" in optimize:
        # Same thing for CSS.
        site.external_manager.register_optimizer(YUICSSOptimizer)

    # Add your own types here!



# #coding:utf-8
# from cactus.tests.integration.s3 import S3TestHTTPConnection, S3IntegrationTestCase
#
#
# class WorkflowTestCase(S3IntegrationTestCase):
#     HTTPConnectionClass = S3TestHTTPConnection
#
#     def test_bucket_name_unset(self):
#         """
#         Test that we prompt the user for the name of the bucket
#         """
#         self.assertEqual(None, self.site.config.get('aws-bucket-name'))
#         self.assertEqual(None, self.site.config.get('aws-bucket-website'))
#
#         bucket_name = "website"  # Need to pick a name that is in our XML list!
#
#         self.site.ui.prompt_normalized = lambda q: bucket_name
#         self.site.upload()
#
#         self.assertEqual(bucket_name, self.site.config.get('aws-bucket-name'))
#         self.assertEqual("{0}.s3-website-us-east-1.amazonaws.com".format(bucket_name),
#             self.site.config.get('aws-bucket-website'))  # See the response we send (US standard).
#
#     def test_no_bucket_create(self):
#         """
#         Test that we prompt for bucket creation if the bucket does not exist
#         """
#         bucket_name = "does-not-exist"
#
#         self.site.config.set('aws-bucket-name', bucket_name)
#         self.assertEqual(None, self.site.config.get('aws-bucket-website'))
#
#         self.site.ui.prompt_yes_no = lambda q: True
#         self.site.upload()
#
#         # Check that we retrieved the list and created a bucket
#         self.assertEqual(4, len(self.connection_factory.requests))
#         list_buckets, create_bucket, enable_website, retrieve_location = self.connection_factory.requests
#
#         self.assertEqual("/", list_buckets.url)
#         self.assertEqual("GET", list_buckets.method)
#
#         self.assertEqual("/{0}/".format(bucket_name), create_bucket.url)
#         self.assertEqual("s3.amazonaws.com".format(bucket_name), create_bucket.connection.host)
#         self.assertEqual("PUT", create_bucket.method)
#
#         self.assertEqual("/{0}/?website".format(bucket_name), enable_website.url)
#         self.assertEqual("PUT", enable_website.method)
#
#         self.assertEqual("/{0}/?location".format(bucket_name), retrieve_location.url)
#         self.assertEqual("GET", retrieve_location.method)
#
#         # Check that we updated our config
#         self.assertEqual("{0}.s3-website-us-east-1.amazonaws.com".format(bucket_name),
#             self.site.config.get('aws-bucket-website'))  # See the response we send (US standard).
#
#     def test_credentials_manager(self):
#         """
#         Test that credentials are saved in the manager
#         """
#         class DummyCredentialsManager(object):
#             saved = False
#
#             def get_credentials(self):
#                 return {"access_key": '123', "secret_key": '456'}
#
#             def save_credentials(self):
#                 self.saved = True
#
#         self.site.deployment_engine.credentials_manager = DummyCredentialsManager()
#         self.site.config.set('aws-bucket-name', 'website')
#         self.site.upload()
#
#         self.assertTrue(self.site.deployment_engine.credentials_manager.saved)
